import os

from . import cfg
from . import io
from . import utils
from . import parser


class Project:
  def __init__(self, args, **kwargs):
    self.args = args
    self.texFileKwargs = kwargs


  def _walk(self, find):
    # walk through directories and find specified stuff
    res = []

    # default to current directory only
    origins = ['.']

    # also walk through all found includedirs for graphics if
    # looking for images
    if find in ('imgs', ):
      for t in self.toplevel():
        for p in t.graphicspath():
          if p not in origins:
            origins.append(p)

    # walk through directories
    hasWarnedDepth = False
    for o in origins:
      for root, dirs, files in os.walk(o, topdown=True):

        # abort tree is too deep
        if root.count(os.sep)-o.count(os.path.sep)-2 > cfg.get('max_directory_depth'):
          io.verb(f'skipping subfolders of {root}')
          if not hasWarnedDepth:
            hasWarnedDepth = True
            io.warn(f'reached max_directory_depth='
                    f'{cfg.get("max_directory_depth")} when recursing '
                    f'through the current directory, ignoring deeper '
                    f'levels')
          dirs.clear()

        # skip git directories
        while True:
          i = [d for d in dirs if d.startswith('.git')]
          if not i:
            break
          dirs.remove(i[0])

        # find and open tex files to detect toplevel
        for f in files:
          if find in ('toplevel', ):
            if f.lower().endswith('.tex'):
              io.verb(f'found tex file {os.path.join(root, f)}')
              file = parser.TexFile(os.path.join(root, f), **self.texFileKwargs)
              if file.isToplevel():
                res.append(file)

          if find in ('imgs', ):
            if (any([f.endswith('.'+(e[1:] if e.startswith('.') else e))
                                  for e in cfg.get('graphics_extensions')])
                # skip pdfs that seem to be builds of tex files
                and not (f.endswith('.pdf')
                          and os.path.exists(os.path.join(root, f[:-4]+'.tex')))):
              io.verb(f'found img {os.path.join(root, f)}')
              file = parser.ImgFile(os.path.join(root, f))
              if file not in res:
                res.append(file)
    return res


  @utils.cacheReturnValue
  def toplevel(self):
    if hasattr(self.args, 'tex_file') and self.args.tex_file:
      return [parser.TexFile(self.args.tex_file, **self.texFileKwargs)]
    else:
      res = self._walk(find='toplevel')
      if res:
        io.verb(f'detected toplevel tex files:',
                *[str(t) for t in res])
      return res


  @utils.cacheReturnValue
  def commonImageDirs(self):
    res = [[p for p in t.graphicspath()] for t in self.toplevel()]
    res = [_p for p in res for _p in p if all([_p in p for p in res])]
    return sorted(list(set(res)))


  @utils.cacheReturnValue
  def allImgFiles(self):
      return sorted(self._walk(find='imgs'))


  @utils.cacheReturnValue
  def allIncludedImgs(self):
    res = []
    for t in self.toplevel():
      for i in t.imgs():
        if i not in res:
          res.append(i)
    return sorted(res)


  @utils.cacheReturnValue
  def unusedIncludedImgs(self):
    res = [f for f in self.allImgFiles()
              if any([f.path.startswith(d) for d in self.commonImageDirs()])]
    for i in self.allIncludedImgs():
      if i in res:
        res.remove(i)
    return sorted(res)


  @utils.cacheReturnValue
  def missingIncludedImgs(self):
    res = [f for f in self.allIncludedImgs()
                          if not f.exists()]
    valid = [f for f in res if not f.containsNewcommandArg()]
    if res != valid:
      io.warn(r'found \includegraphcis{} call with #n in argument, cannot '
              r'check if image exists for such calls')
    return sorted(valid)


  @utils.cacheReturnValue
  def allBibs(self):
    res = []
    for t in self.toplevel():
      for b in t.bibs():
        if b not in res:
          res.append(b)
    return res


  @utils.cacheReturnValue
  def allCitesInBibs(self):
    res = []
    for t in self.toplevel():
      for b in t.bibs():
        for c in b.cites():
          if c not in res:
            res.append(c)
    return sorted(res)


  @utils.cacheReturnValue
  def allCitesInTex(self):
    res = []
    for t in self.toplevel():
      for c in t.cites():
        if c not in res:
          res.append(c)
    return sorted(res)


  @utils.cacheReturnValue
  def unusedCites(self):
    res = self.allCitesInBibs()
    for c in self.allCitesInTex():
      if c in res:
        res.remove(c)
    return sorted(res)


  @utils.cacheReturnValue
  def missingCites(self):
    res = []
    for t in self.toplevel():
      for c in t.missingCites():
        if c.isHealthy and c not in res:
          res.append(c)
    return res


  @utils.cacheReturnValue
  def lint(self):
    visited = []
    for t in self.toplevel():
      for l in t.lint(visited=visited):
        yield l



from ..parser import journal
from .common import *

def main(args):
  if args.list_suspicious:
    journal.printSuspicious()
  else:
    m = journal.find(*args.name)
    if not m:
      io.info('no journals found')
    else:
      longestAbbr = max([len(_m[0]) for _m in m])
      for abbr, name in m:
        if args.full:
          io.info(name)
        elif args.abbreviated:
          io.info(abbr)
        else:
          abbr += '": '
          io.info(f'"{abbr:<{longestAbbr+3}}"{name}"')



import os
import re

from .. import cfg
from .. import io
from .. import utils
from . import common

from .cite import Cite, FORBIDDEN_KEY_CHARS


class BibFile:
  def __init__(self, fname):
    self.fname = os.path.normpath(fname)
    self.path = None
    if self.exists():
      self.path = self.exists()
      self._exists = True


  @utils.cacheReturnValue
  def exists(self):
    d = os.path.dirname(self.fname) or '.'
    if not os.path.isdir(d):
      return False
    for ext in ['']+list(cfg.get('bibtex_extensions')):
      candidate = utils.replaceSuffix(self.fname, ext)
      #io.dbg(f'trying candidate {candidate}')
      if os.path.isfile(candidate):
        return candidate
    return False


  @utils.cacheReturnValue
  def content(self):
    if self.path:
      with open(self.path, 'r') as f:
        return '\n'.join([l.rstrip('\n') for l in f])
    return ''


  @utils.cacheReturnValue
  def cites(self):
    res = []

    # parsing whole file with a characterwise state machine is the only
    # safe way to go...
    state = 'comment'
    currentSection = None
    currentKey = None
    currentItem = None
    currentValue = None
    currentValueParseInfo = None
    currentLine = 1
    previousBackslash = False
    braceDepth = None
    quoteDepth = None
    c = None

    def raiseErr(msg):
      nonlocal currentLine, state, c
      if cfg.get('debug'):
        allLines = self.content().split('\n')
        lines = allLines[max(0, currentLine-5):min(len(allLines), currentLine+5)]
        pretty = [('=> ' if i+1+max(0, currentLine-5) == currentLine else '>  ')+l
                    for i, l in enumerate(lines)]
      raise RuntimeError(f'in bib file {self.path} (line {currentLine}):\n'
                         +('\n'.join(pretty)+'\n'
                           +f'parser state "{state}", current char {repr(c)}\n'
                                                  if cfg.get('debug') else '')
                         +f'parsing error, {msg}')

    def _assert(expr, msg='unexpected error'):
      if not expr:
        raiseErr(msg)

    # function to call whenever an item-value pair is completed
    def flushItemValue():
      nonlocal currentItem, currentValue, currentValueParseInfo, braceDepth, quoteDepth
      currentItem = currentItem.strip()
      _assert(currentItem,
              'found empty item name')
      _assert(all([(c.isalpha() or c in '-_') for c in currentItem]),
              f'found invalid item name "{currentItem.strip()}"')
      _assert(len(res))
      if res[-1].fields is None:
        res[-1].fields = {}
      if res[-1].fieldsParseInfo is None:
        res[-1].fieldsParseInfo = {}
      if currentItem in res[-1].fields:
        raiseErr(f'duplicate item "{currentItem}"')

      # strip content between outermost braces/quotes
      if ('opens' in currentValueParseInfo
          and 'closes' in currentValueParseInfo
          and len(_opens := currentValueParseInfo['opens'])
          and len(_closes := currentValueParseInfo['closes'])):
        # extract content between outermost curly braces
        _o, _c = _opens[0], _closes[-1]
        midPart = currentValue[_o+1:_c]

        # shift last closing index according to removed spaces
        _closes[-1] -= len(midPart) - len(midPart.strip())

        # replace currentValue with stripped variant
        currentValue = (currentValue[:_o+1]
                        + midPart.strip()
                        + currentValue[_c:]
                       ).strip()

      res[-1].fields[currentItem] = currentValue.strip()
      res[-1].fieldsParseInfo[currentItem] = currentValueParseInfo
      #io.dbg(f'created new field for citation {res[-1].key}',
      #       f'item: {currentItem}, value: {currentValue}',
      #       f'parseInfo: {currentValueParseInfo}',
      #       f'in {self.path}:{currentLine}')
      currentValue = None
      currentValueParseInfo = None
      currentItem = None
      braceDepth = None
      quoteDepth = None

    for c in self.content():
      # @ sign while in "comment" -> transition to "section"
      if c == '@' and state == 'comment':
        _assert(currentSection is None)
        currentSection = ''
        state = 'section'
        #io.dbg(f'found start of new bib section',
        #       f'in {self.path}:{currentLine}')

      # skip all characters outside of sections
      elif state == 'comment':
        pass

      # { sign while in "section" -> transition to "key"
      elif c == '{' and state == 'section':
        _assert(currentSection is not None)
        _assert(currentSection.strip().isalpha(),
                f'invalid section name {repr(currentSection)}')
        _assert(currentKey is None)

        # special case section=comment: immediately transition
        # back to comment state
        if currentSection.strip().lower() == 'comment':
          currentSection = None
          state = 'comment'
        else:
          currentKey = ''
          state = 'key'

      # record any other character
      elif state == 'section':
        _assert(currentSection is not None)
        currentSection += c

      # , sign while in "key" -> transition to "item" and flush key and section
      elif c == ',' and state == 'key':
        _assert(currentSection is not None
                  and currentKey is not None)
        _assert(currentSection,
                'empty section name')
        _assert(currentKey,
                'empty key')
        res.append(Cite(currentKey,
                        bibs=[self],
                        section=currentSection))
        #io.dbg(f'created new citation entry with key {currentKey}',
        #       f'and type {currentSection}',
        #       f'in {self.path}:{currentLine}')
        currentKey = None
        currentSection = None

        _assert(currentItem is None)
        currentItem = ''
        state = 'item'

      # finding closing curly brace in key means bibtex entry without any items,
      # flush key and section and transition to comment state
      elif c == '}' and state == 'key':
        _assert(currentItem is None
                  and currentValue is None)
        res.append(Cite(currentKey,
                        bibs=[self],
                        section=currentSection))
        #io.dbg(f'created new empty citation entry with key {currentKey}',
        #       f'and type {currentSection}',
        #       f'in {self.path}:{currentLine}')
        currentKey = None
        currentSection = None
        state = 'comment'

      # any other character records key
      elif state == 'key':
        _assert(currentKey is not None)
        if len(currentKey.strip()):
          _assert(c not in ' \n\t',
                  f'unexpected end of citation key "{currentKey}", expected ","')
        _assert(c not in FORBIDDEN_KEY_CHARS,
                f'forbidden charcter in citation key: {repr(c)}')
        currentKey += c

      # = sign while in "item" -> transition to "value" and flush key and section
      elif c == '=' and state == 'item':
        state = 'value'
        _assert(currentSection is None
                  and currentKey is None)
        _assert(braceDepth is None
                  and quoteDepth is None
                  and currentValue is None)
        braceDepth = 0
        quoteDepth = 0
        currentValue = ''
        currentValueParseInfo = dict(opens=[], closes=[])

      # if a closing curly brace is encountered in item state and the current
      # item is empty, this means the section has ended, then flush everything
      # and transition to comment state
      elif c == '}' and state == 'item':
        _assert(currentItem.strip() == '')
        currentItem = None
        _assert(currentSection is None
                  and currentKey is None
                  and currentValue is None)
        state = 'comment'

      # record all other characters as item name
      elif state == 'item':
        _assert(currentItem is not None)
        _assert(currentItem.strip() == ''
                  or all([(c.isalpha()
                              or c.strip() == ''
                              or c in '-_') for c in currentItem]),
                f'found invalid item name starting with "{currentItem.strip()}"')
        currentItem += c

      # nice summary of how {, } and " delimiters are parsed by bibtex:
      # https://tex.stackexchange.com/questions/109064/is-there-a-difference-between-and-in-bibtex
      #
      # value parsing in short:
      # * do not consume curly braces, always error on unmatched curly braces.
      # * do not consume double-quotes only count depth and error on unmatched
      #   double-quotes only if not within {}
      # * treat single quotes as normal character
      # * backslash escapes {, }, but still count depth to guarantee matching
      # * backslash escapes double-quotes
      # * if any different character follows backslash, treat backslash as
      #   normal character
      elif state == 'value':

        # if previous character was a backslash, modify current character
        # such that it will be treated as normal character
        if previousBackslash:
          c = '\\'+c
          previousBackslash = False

        # do not handle backslash now but only set previousBackslash flag
        if c == '\\':
          previousBackslash = True

        # { sign while in "value" -> increment braceDepth
        elif c == '{':
          _assert(braceDepth is not None)
          braceDepth += 1
          _assert(currentValue is not None)
          currentValue = currentValue.lstrip()
          currentValueParseInfo['opens'].append(len(currentValue))
          currentValue += c

        # a seemingly unmatched } in value state means that citation is
        # complete, flush everything and transition to "comment" state
        elif c == '}' and braceDepth == 0 and quoteDepth == 0:
          flushItemValue()
          _assert(currentSection is None
                    and currentKey is None
                    and currentItem is None
                    and currentValue is None)
          state = 'comment'

        # } sign while in "value" -> decrement braceDepth
        elif c == '}':
          _assert(braceDepth is not None and braceDepth > 0,
                  'found unmatched "}"')
          braceDepth -= 1
          _assert(currentValue is not None)
          currentValue = currentValue.lstrip()
          currentValueParseInfo['closes'].append(len(currentValue))
          currentValue += c

        # " sign while in "value" -> toggle quote depth if not within
        # curly braces, add to currentValue anyways
        elif c == '"' and state == 'value' and braceDepth == 0:
          _assert(quoteDepth is not None)
          currentValue = currentValue.lstrip()
          if quoteDepth:
            currentValueParseInfo['closes'].append(len(currentValue))
          else:
            currentValueParseInfo['opens'].append(len(currentValue))
          quoteDepth = (1 if quoteDepth == 0 else 0)
          currentValue += c

        # add escaped curly brace to content and also increment depth
        elif c == r'\{':
          _assert(braceDepth is not None)
          braceDepth += 1
          currentValue += c

        # add escaped curly brace to content and also decrement depth
        elif c == r'\}':
          _assert(braceDepth is not None and braceDepth > 0,
                  'found unmatched "}"')
          braceDepth -= 1
          currentValue += c

        # , sign with balanced braces and quotes flushes current item value
        # pair and leaves value mode
        elif c == ',' and braceDepth == 0 and quoteDepth == 0:
          flushItemValue()
          currentItem = ''
          state = 'item'

        # any other character is just added to content
        else:
          currentValue += c

      else:
        _assert(state == 'comment',
                f'unexpected character {repr(c)}')

      # increase line counter on newlines no matter what state
      if c == '\n':
        currentLine += 1

    _assert(currentSection is None
                and currentKey is None
                and currentItem is None
                and currentValue is None
                and previousBackslash is False
                and braceDepth is None
                and quoteDepth is None
                and state == 'comment',
            'citation section was started but never ended,\n'
            'probably due to unmatched curly braces')

    return res


  def _duplicates(self):
    byKey, byGeneratedKey, byCompareAuthorTitle = {}, {}, {}
    for cite in self.cites():
      for d, k in zip([byKey, byGeneratedKey, byCompareAuthorTitle],
                      [cite.key, cite.makeKey(), cite.compareAuthorTitle()]):
        if k not in d.keys():
          d[k] = []
        d[k].append(cite)
    return [ [c for c in d.values() if len(c)>1]
                    for d in (byKey, byGeneratedKey, byCompareAuthorTitle) ]

  
  def duplicatesKeys(self):
    return self._duplicates()[0]


  def duplicatesGeneratedKeys(self):
    return self._duplicates()[1]


  def duplicatesAuthorTitle(self):
    return self._duplicates()[2]


  def addCites(self, cites):
    if cites:
      with open(self.path, 'a') as f:
        f.write('\n'+'\n\n'.join([c.pretty() for c in cites])+'\n')
      self._cites = None
      self.cites()


  def setCites(self, cites):
    newContent = '\n\n'.join([c.pretty() for c in cites])+'\n'
    if newContent != self.content:
      with open(self.path, 'w') as f:
        f.write(newContent)
      self._cites = None
      self.cites()


  def rewrite(self):
    self.setCites(self.cites())


  def create(self):
    if not self.exists():
      fname = self.fname
      with open(utils.replaceSuffix(fname, cfg.get('bibtex_extensions')[0]), 'a') as f:
        pass
      self.__dict__ = {}
      self.__init__(fname)


  def fromRis(key, risFile):
    entries = dict()
    entries['authors']=list()

    with open(risFile) as f:
      for line in f:
        if re.match("PY",line):
          entries['year'] = line[6:10]
        elif re.match("AU",line):
          entries['authors'].append(line[6:-1])
        elif re.match("VL",line):
          entries['volume'] = line[6:-1]
        elif re.match("TI",line):
          entries['title'] = line[6:-1]
        elif re.match("T1",line):
          entries['title'] = line[6:-1]
        elif re.match("JO",line):
          entries['journal'] = line[6:-1]
        elif re.match("IS",line):
          entries['number'] = line[6:-1]
        elif re.match("SP",line):
          entries['startpage'] = line[6:-1]
        elif re.match("EP",line):
          entries['endpage'] = line[6:-1]
        elif re.match("SN",line):
          entries['isbn'] = line[6:-1]
        elif re.match("AB",line):
          entries['abstract'] = line[6:-1]
        elif re.match("UR",line):
          entries['url'] = line[6:-1]

    bibStr = ('@article{' + key + ',\n'
                  + 'author={' + (" and ".join(entries['authors'])) + '},\n'
                  + 'year={' + entries['year'] + '},\n'
                  + 'title={' + entries['title'] + '},\n'
                  + ('abstract' in entries) * ('abstract = {' + entries.get('abstract', '') + '},\n')
                  + ('journal' in entries) * ('journal={' + entries.get('journal', '') + '},\n')
                  + ('volume' in entries) * ('volume={' + entries.get('volume', '') + '},\n')
                  + ('startpage' in entries and 'endpage' in entries)
                      * ('pages={' + entries.get('startpage', '')
                          + "--" + entries.get('endpage', '') + '},\n')
                  + 'url={'+entries['url']+'},\n'
                  + '}\n')
    io.dbg('converted ris to bib: ', bibStr)
    res = BibFile('')
    res._content = bibStr
    return res


  def lint(self, visited=[]):
    reported = []

    # report duplicates
    for duplicates in self.duplicatesKeys():
      if duplicates not in reported:
        reported.append(duplicates)
        yield (self.path, '?', f'found bib entries with duplicate '
                  f'key {duplicates[0].key}')

    for duplicates in self.duplicatesGeneratedKeys():
      if duplicates not in reported:
        reported.append(duplicates)
        yield (self.path, '?', f'found bib entries with conflicting generated '
                  f'keys:\n{", ".join([d.key for d in duplicates])}')

    for duplicates in self.duplicatesAuthorTitle():
      if duplicates not in reported:
        reported.append(duplicates)
        yield (self.path, '?', f'found bib entries with identical authors and title:\n'
                  f'{", ".join([d.key for d in duplicates])}')



import logging
import logging.handlers as handlers
import os
import traceback
import appdirs
from numpy import log10

from . import cfg

_LOG_DIR             = appdirs.user_log_dir('paperman')
_LOGFILE_NAME        = 'paperman.log'
_ROTATE_DIR          = 'oldlogs'
_IS_INIT             = False

isVerbose            = False

def _logger():
  return logging.getLogger('paperman')


def _init():
  global _IS_INIT, _LOG_DIR
  if not type(_LOG_DIR) is str:
    _LOG_DIR = _LOG_DIR()
  os.makedirs(_LOG_DIR, exist_ok=True)
  for oldlog in [f for f in os.listdir(_LOG_DIR)
                    if f != _LOGFILE_NAME and f.startswith(_LOGFILE_NAME)]:
    os.makedirs(_LOG_DIR+'/'+_ROTATE_DIR, exist_ok=True)
    os.rename(_LOG_DIR+'/'+oldlog, _LOG_DIR+'/'+_ROTATE_DIR+'/'+oldlog)

  if not _IS_INIT:
    h = handlers.TimedRotatingFileHandler(_LOG_DIR+'/'+_LOGFILE_NAME, when='W6')
    h.setFormatter(logging.Formatter('%(asctime)s %(levelname)s: %(message)s',
                                     datefmt='%d.%m. %I:%M:%S'))
    l = _logger()
    l.addHandler(h)
    if cfg.get('debug'):
      l.setLevel(logging.DEBUG)
    else:
      l.setLevel(logging.INFO)
    _IS_INIT = True


def setLogfile(name):
  global _IS_INIT, _LOGFILE_NAME
  name = str(name)
  if '/' in name:
    raise ValueError(f'logfile name must not contain a slash "/" : {name}')
  if not name.endswith('.log'):
    name += '.log'
  _logger().handlers.clear()
  _IS_INIT = False
  _LOGFILE_NAME = name
  _init()


def _indentMsg(msg):
  ls = [l for l in '\n'.join([str(l) for l in msg]).split('\n') if l.strip()]
  if len(ls) == 0:
    return ''
  if len(ls) == 1:
    return ls[0]
  elif len(ls) == 2:
    return ls[0]+'\n'+' '*2+'\ '+ls[1]
  else:
    return ('\n'+' '*2+'| ').join(ls[:-1])+'\n'+' '*2+'\ '+ls[-1]


def err(*msg, logOnly=False):
  _init()
  msg = _indentMsg(msg)
  _logger().error(msg)
  if not logOnly:
    print('error:', msg)
    if '\n' in msg:
      print()


def formatErr(*msg):
  return 'error: '+_indentMsg(msg)


def warn(*msg, logOnly=False):
  _init()
  msg = _indentMsg(msg)
  _logger().warning(msg)
  if not logOnly:
    print('warning:', msg)
    if '\n' in msg:
      print()


def info(*msg, logOnly=False, noNewLine=False):
  _init()
  msg = _indentMsg(msg)
  _logger().info(msg)
  if not logOnly:
    print(msg)
    if '\n' in msg and not noNewLine:
      print()


def raw(*msg):
  _init()
  _logger().info(_indentMsg(msg))
  print(*msg)


def verb(*msg, logOnly=False):
  if isVerbose:
    info(*msg, logOnly=logOnly)
  elif cfg.get('debug'):
    dbg(*msg, logOnly=logOnly)


def dbg(*msg, logOnly=False):
  _init()
  msg = _indentMsg(msg)
  _logger().debug(msg)
  if not logOnly and cfg.get('debug'):
    print('debug:', msg)
    if '\n' in msg:
      print()


def conf(*msg, default=None):
  _init()
  msg = _indentMsg(msg)
  if default is None:
    msg += ' (y/n)'
  elif default:
    msg += ' (Y/n)'
  else:
    msg += ' (y/N)'
  res = None
  while res is None:
    _logger().info(msg)
    print(msg, end=' ')
    i = input()
    _logger().info('user input: "'+str(i)+'"')
    if '\n' in msg:
      print()
    if i.strip() == '':
      res = default
    elif i.strip().lower() in ('y', 'yes'):
      res = True
    elif i.strip().lower() in ('n', 'no'):
      res = False
    if res is None:
      info('invalid input, expecting (y)es or (n)o\n')
  return res


def select(msg, options, default=None):
  _msg = (msg+'\n'
           +'\n'.join([
               f'{"" if default is None else ("default: " if i+1==default else " "*9)}'
               f'({i+1:{int(log10(len(options))+1)}d})'
               f' {str(o)}' for i, o in enumerate(options)]))
  res = None
  while res is None:
    info(_msg, noNewLine=True)
    i = input()
    _logger().info('user input: "'+str(i)+'"')
    if '\n' in _msg:
      print()
    if i.strip() == '':
      res = default
    else:
      try:
        res = int(i)
        if res < 1 or res > len(options):
          raise ValueError('value out of range')
      except ValueError:
        res = None
    if res is None:
      info(f'invalid input, expecting number between 1 and {len(options)}\n')
  return res-1, options[res-1]


def startup():
  _init()
  msg = _indentMsg([f'starting paperman v{__version__}'])
  info('', logOnly=True)
  info('-'*len(msg), logOnly=True)
  info(msg, logOnly=True)
  if cfg.get('debug'):
    print('debug:', msg)


def coercedStacktrace(lines):
  excLines = traceback.format_exc().split('\n')
  if len(excLines) <= lines:
    return '\n'.join(excLines)
  else:
    return (f'skipped {len(excLines)-lines} stack trace lines...\n...\n'
            +'\n'.join(excLines[-lines:]))



import re
import time
import shutil

from .. import parser
from . common import *


def main(args):
  # check of library path is set
  libraryPath = os.path.expanduser(cfg.get('library_path'))
  if not libraryPath:
    io.err('library_path is not set in config file')
    return

  # check if collect paths exist
  collectPaths = args.paths or cfg.get('library_collect_paths')
  if collectPaths:
    io.verb('colling bib-pdf pairs from the follwoing paths:', *collectPaths)
  else:
    io.err('no collect paths specified, either',
           'set "library_collect_paths" in the config or',
           'pass paths as argument to "paperman collect"')
    return

  # define routine to use for error reporting
  lastErrStr = ''
  def onError(file, *err):
    nonlocal lastErrStr
    errStr = io.formatErr(f'failed to collect "{file}":', *err)
    if lastErrStr.lower().strip() != errStr.lower().strip():
      if args.err_to_file:
        with open(os.path.join(os.path.dirname(file),
                               'paperman-collect-error.txt'), 'w') as f:
          f.write(errStr)
      io.err(f'failed to collect "{file}":', *err)
      lastErrStr = errStr

  # mainloop
  while True:
    try:
      t0 = time.time()
      for p in collectPaths:
        p = os.path.expanduser(p)

        pdfs = [os.path.join(p, f) for f in os.listdir(p)
                          if os.path.isfile(os.path.join(p, f))
                              and f.lower().endswith('.pdf')]

        bibs = [os.path.join(p, f) for f in os.listdir(p)
                          if os.path.isfile(os.path.join(p, f))
                              and any([f.lower().endswith(e)
                                  for e in list(cfg.get('bibtex_extensions'))
                                                  +['.ris']])]

        io.verb('in folder ', p,
                'found pdf candidates: ', *(pdfs or ['-none-']),
                'found bib candidates: ', *(bibs or ['-none-']))

        if len(pdfs) == 1 and len(bibs) == 1:
          pdfPath = pdfs[0]
          bibPath = bibs[0]

          # load and parse bib file (retry five times in case bib file was not
          # completely written when it was initially discovered)
          cites = []
          for _ in range(5):
            try:
              if bibPath.lower().endswith('.ris'):
                r = parser.BibFile.fromRis('key', bibPath)
                io.dbg(r)
                cites = r.cites()
              else:
                cites = parser.BibFile(bibPath).cites()
              if len(cites) != 1:
                raise ValueError('invalid number of citations in bib file')
              break
            except Exception:
              time.sleep(.1)

          io.dbg('cites:', cites)
          if len(cites) == 1:
            cite = cites[0]

            # run pretty method to clean up fields and trigger potential reformatting questions
            cite.pretty()

            # generate key and filename from paper title
            cite.key = cite.makeKey()
            libraryDir = os.path.join(libraryPath,
                                      time.strftime(cfg.get('library_folder_pattern')))
            target = os.path.join(libraryDir, cite.key)
            os.makedirs(os.path.dirname(target), exist_ok=True)
            if not any([f.startswith(target)
                            for f in os.listdir(os.path.dirname(target))]):
              targetName = target+'.'+cfg.get('bibtex_extensions')[0]
              try:
                with open(targetName, 'w') as f:
                  f.write(cite.pretty()+'\n')
              except Exception:
                if os.path.exists(targetName):
                  os.remove(targetName)
                raise
              os.remove(bibPath)
              #io.info(f'moved "{bibPath}" -> "{target}.{cfg.get("bibtex_extensions")[0]}"')
              shutil.move(pdfPath, target+'.pdf')
              #io.info(f'moved "{pdfPath}" -> "{target}.pdf"')
              io.info(f'imported "{cite.key}"')

              # reset lastErr memory
              lastErrStr = ''
            else:
              onError(pdfPath, 'failed to move files to library, target',
                      '"'+target+'"', 'already exists')

          else:
            onError(pdfPath, f'found unexpected citation count ({len(cites)}) in bib file')

        # situation that naturally occurs, do not print error
        elif (len(pdfs)==1 and len(bibs)==0) or (len(pdfs)==0 and len(bibs)==1):
          pass

        elif len(pdfs) or len(bibs):
          onError(p, f'found {len(pdfs)} pdf candidates and {len(bibs)} bib',
                     f'need to have exactly one of each to collect')

      # if watch option is not set run mainloop only once
      if not args.watch:
        break

      # calculate sleep to not consume more than 1% of cpu-load, but also not
      # faster less than 0.1s or more than 5s
      sleepTime = min(5, max(.1, 1e2*(time.time()-t0)))
      io.verb(f'sleeping for {sleepTime:.1f}s')
      time.sleep(sleepTime)

    except KeyboardInterrupt:
      io.info('received ctrl+c interrupt, exiting...')
      break
    except:
      if not args.watch:
        raise
      io.err('unexpected error:', io.coercedStacktrace(30))
      time.sleep(5)



import subprocess
import shutil
import time

from .common import *

def main(args):
  # extract sync path from config or from argument list
  if args.sync_path:
    syncPath = args.sync_path
  else:
    syncPath = os.path.expanduser(cfg.get('library_sync_device'))
  if not syncPath:
    io.err('library_sync_device config is not set up')
    return

  # clean path
  while syncPath.endswith('/'):
    syncPath = syncPath[:-1]

  # test if sync path is among mounted paths
  mounts = subprocess.run(['mount'], capture_output=True).stdout.decode()
  if syncPath not in mounts:
    io.err(f'sync device at {syncPath} does not seem to be mounted')
    return

  # try to place busy file
  try:
    with open(syncPath+'/paperman-sync-busy', 'w') as f:
      pass
  except KeyboardInterrupt:
    raise
  except:
    io.err(f'failed to place busy file at {syncPath}, do you have write access?')
    return

  # init flag
  nothingDone = True

  # walk through library pdfs and upload missing ones
  libPath = os.path.expanduser(cfg.get('library_path'))
  additionalPaths = [os.path.expanduser(p) for p in cfg.get('library_sync_additional_paths')]
  for _path in [libPath]+additionalPaths:
    for root, dirs, files in os.walk(_path, topdown=True):
      dirs[:] = [d for d in dirs if not d.startswith('.')]
      files[:] = [f for f in files if not f.startswith('.')]

      for f in files:
        if _path == libPath:
          relPath = root[len(_path):]
        else:
          relPath = root[_path.rfind('/')+1:]
        if relPath.startswith('/'):
          relPath = relPath[1:]

        if f.endswith('.pdf') and not relPath.startswith('annotated'):

          # check if file looks valid and is not too old
          s = os.stat(root+'/'+f)
          if (s.st_size > 0 and
              time.time()-s.st_mtime < cfg.get('library_sync_max_age')):

            # create target path
            target = f'{syncPath}/{relPath}/{f}'
            if not os.path.exists(target):
              io.info(f'uploading {relPath}/{f}')
              os.makedirs(os.path.dirname(target), exist_ok=True)
              shutil.copy(root+'/'+f, target)
              nothingDone = False

  # copying files that changed on the device to annotated folder
  for root, dirs, files in os.walk(syncPath, topdown=True):
    dirs[:] = [d for d in dirs if not d.startswith('.')]
    files[:] = [f for f in files if not f.startswith('.')]

    for f in files:
      # check if file also exists in library but has different size
      relPath = root[len(syncPath):]
      if relPath.startswith('/'):
        relPath = relPath[1:]

      if f.endswith('.pdf'):
        # skip folder names beginning with a dot
        if not relPath.startswith('.'):

          # go through all possible local path candidates, if any of them
          # exists and has the same file size, do not copy it
          localCandidates = ([f'{libPath}/{relPath}/{f}']
                            +[f'{_p[:_p.rfind("/")]}/{relPath}/{f}'
                                                  for _p in additionalPaths])
          if not any([os.path.exists(l)
                        and os.stat(l).st_size == os.stat(root+'/'+f).st_size
                                for l in localCandidates]):

            # create target path, if it exists, modify target path
            # and copy only if size of existing file differs
            target = f'{libPath}/annotated/{relPath}/{f}'
            origTarget = target
            existing = None
            i = 0
            while os.path.exists(target):
              i += 1
              existing = target
              target = ('.'.join(origTarget.split('.')[:-1])
                            +f'-v{i:02d}.'
                            +origTarget.split('.')[-1])

            # copy file only if last version of annotated file has different
            # size than current version of annotated file (or if target does
            # not exist)
            if (existing is None or
                  os.stat(existing).st_size != os.stat(root+'/'+f).st_size):
              io.info(f'downloading annotated {relPath}/{f}')
              os.makedirs(os.path.dirname(target), exist_ok=True)
              shutil.copy(root+'/'+f, target)
              nothingDone = False

  # indicate that something happened even if nothing was copied
  if nothingDone:
    io.info('nothing to do')

  # remove busy file
  os.remove(syncPath+'/paperman-sync-busy')



from . import io
from . import cfg
from . import utils


# try to extract version info
try:
  import importlib.metadata
  __version__ = importlib.metadata.version('paperman')
except:
  try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution('paperman').version
  except:
    __version__ = '???'

# set version attribute for submodules that need it
for m in [io, ]:
  m.__version__ = __version__



import re
import requests
import cloudscraper
import time
import unidecode

from .. import io
from .. import cfg
from .. import utils

FORBIDDEN_KEY_CHARS = r'''#'",=(){}%~\ '''

@utils.cacheReturnValue
def hasNetwork():
  for _ in range(5):
    if _hasNetwork():
      return True
    time.sleep(1)
  return _hasNetwork()

def _hasNetwork():
  t0 = time.time()
  try:
    requests.get('https://pypi.org/project/paperman/', timeout=1)
    result = True
  except KeyboardInterrupt:
    raise
  except:
    io.warn('not connected to the internet, cannot check citation urls '
            'and dois')
    result = False

  # begin of workaround
  #  if ipv6 is misconfigured or not availabe requests.get(...) works
  #  but is extremely slow, because it only falls back to ipv4 after
  #  the timeout. This workaround (dirtily) detects misconfigured ipv6
  #  by measuring the time that requests.get(...) took and disables
  #  future use of ipv6 connections if it took longer than the
  #  timeout.
  if result and time.time()-t0 > 1:
    io.verb('disabling ipv6 connections')
    requests.packages.urllib3.util.connection.HAS_IPV6 = False
  # end of workaround

  return result


def isDoiValid(doi):
  io.dbg(f'verifying doi {doi}...')
  err = _isDoiOrUrlValid(doi, 'z_verified_dois',
                          prefix='https://doi.org/')
  if not err:
    return ''
  return f'invalid doi: {err}'


def isUrlValid(url):
  io.dbg(f'verifying url {url}...')
  err = _isDoiOrUrlValid(url, 'z_verified_urls')
  if not err:
    return ''
  return f'invalid url: {err}'


_cloudscraper = None

def _getScraper():
  global _cloudscraper
  if _cloudscraper is None:
    _cloudscraper = cloudscraper.create_scraper()
  return _cloudscraper


def _isDoiOrUrlValid(url, validCfgListKey, prefix=''):
  url = url.strip()
  try:
    verified = [(t, u) for t, u in cfg.get(validCfgListKey)
                      if time.time()-t < 30*24*60*60*cfg.get('bib_repair',
                              'doi_and_url_verification_max_age_months')]
  except KeyboardInterrupt:
    raise
  except:
    verified = []

  # check if in list
  if url in [u for t, u in verified]:
    result = ''

  # try to fetch urls
  else:
    result = ''
    #io.dbg(f'requesting {prefix+url}')
    r = None
    if hasNetwork():
      io.dbg(f'requesting url {prefix+url}')
      try:
        r = requests.get(prefix+url, timeout=5)
        if r.status_code//10 != 20:
          if 'cloudflare' in r.text.lower():
            raise ValueError('cloudflare protection')
          raise ValueError(f'status code is {r.status_code}')
      except KeyboardInterrupt:
        raise
      except Exception as _e:
        try:
          r = _getScraper().get(prefix+url, timeout=5)
          if r.status_code//10 != 20:
            raise ValueError(f'status code is {r.status_code}')
        except KeyboardInterrupt:
          raise
        except Exception as e:
          result = 'requests: '+str(_e)+', cloudscraper: '+str(e)

  if not result:
    verified.append((time.time(), url))

  cfg.set(validCfgListKey, verified)
  return result


_printedCloudflareWarning = False

def _printCloudflareWarning():
  global _printedCloudflareWarning
  if not _printedCloudflareWarning:
    io.warn(f'cloudflare protection prevented checking some URLs and/or DOIs')
    _printedCloudflareWarning = True


def shouldBeProtected(fullString, string, askProtect=True):
  protectedList = [s.strip()
                        for s in cfg.get('z_bib_words_protect_capitalization')]
  antiProtectedList = [s.strip()
                        for s in cfg.get('z_bib_words_dont_protect_capitalization')]
  string = string.strip()
  if string in antiProtectedList:
    return False
  if string in protectedList:
    return True
  decision = True
  if askProtect:
    decision = io.conf(f'in bib line',
                       f'"{fullString}"',
                       f'should the capitalization of ',
                       f'"{string}"',
                       f'be protected in the bibliography?', default=True)
  if decision:
    if io.conf(f'should protected capitalization of',
               f'"{string}"',
               f'be saved permanently?', default=True):
      protectedList.append(string)
      cfg.set('z_bib_words_protect_capitalization',
              sorted(list(set(protectedList))))
  else:
    if io.conf(f'should never capitalizing',
               f'"{string}"',
               f'be saved permanently?', default=False):
      antiProtectedList.append(string)
      cfg.set('z_bib_words_dont_protect_capitalization',
              sorted(list(set(antiProtectedList))))
  return decision


class Cite:
  def __init__(self, key, bibs=None, section=None, fields=None):
    self.key = key.strip()
    self.bibs = bibs
    self.section = (None if section is None else section.strip().lower())
    self.fields = fields
    self.fieldsParseInfo = None


  def __repr__(self):
    return f'citation {self.key}'+('' if self.isHealthy() else ' (unhealthy)')


  def __eq__(self, cite):
    return self.key == cite.key


  def compareAuthorTitle(self):
    cmp = lambda a: re.sub(r'[^a-z0-9]+', '', a.lower()).replace('and', '')
    return 'cmp'+cmp(self.get('author', ''))+cmp(self.get('title', ''))


  def authorAndTitleEqual(self, cite):
    return self.compareAuthorTitle() == cite.compareAuthorTitle()


  def __gt__(self, cite):
    return self.key > cite.key


  def __lt__(self, cite):
    return self.key < cite.key


  def __getitem__(self, item):
    _, val = self._findFieldKeyValuePair(item)
    return val


  def get(self, item, default=None):
    res = self[item]
    if res is None:
      return default
    return res


  def __setitem__(self, item, value):
    k, val = self._findFieldKeyValuePair(item)
    if k is not None:
      # pop item element
      self.fields.pop(k)

      # pop parse info as well, which is useless after setting a new value
      if (self.fieldsParseInfo is not None
              and k in self.fieldsParseInfo):
        self.fieldsParseInfo.pop(k)

    # set new value
    value = str(value)
    if value[0] != '{' and value[-1] != '}':
      value = '{'+str(value)+'}'
    io.dbg(f'setting new field self.fields[{item.lower()}] = {repr(value)}',
           f'self.fieldsParseInfo[{item.lower()}] = {dict(opens=[0], closes=[len(value)-1])}')
    self.fields[item.lower()] = value
    self.fieldsParseInfo[item.lower()] = {'opens': [0],
                                          'closes': [len(value)-1]}


  # return item from self.items with case insensitive matching
  def _findFieldKeyValuePair(self, item):
    if self.fields is None:
      return None, None

    keys = list(self.fields.keys())
    lkeys = [k.lower() for k in keys]
    if item.lower() not in lkeys:
      return None, None

    k = keys[lkeys.index(item.lower())]

    # if limited by { and } or ", strip those
    f = self.fields[k]
    while (( f.strip().startswith('{') 
              and f.strip().endswith('}') )
        or ( f.strip().startswith('"') 
              and f.strip().endswith('"') ) ):
      f = f.strip()[1:-1]

    # return key and result
    return k, f


  def insertNonexistingItems(self, cite):
    lkeys = [k.lower() for k in self.fields.keys()]
    for k, v in cite.fields.items():
      if k not in lkeys:
        parseInfo = {}
        if k in v.fieldsParseInfo:
          parseInfo = v.fieldsParseInfo[k]
        _k = k
        if cfg.get('bib_repair', 'make_all_items_lowercase'):
          _k = k.lower()
        io.dbg(f'inserted field {_v}={v[:10]}{"" if len(v)<10 else "..."} into citation')
        self.fields[_k] = v
        self.fieldsParseInfo[_k] = parseInfo


  @utils.cacheReturnValue
  def isHealthy(self):
    # list of forbidden characters in bibtex keys found here:
    # https://tex.stackexchange.com/questions/408530/what-characters-are-allowed-to-use-as-delimiters-for-bibtex-keys
    return all([c not in self.key for c in FORBIDDEN_KEY_CHARS])


  @utils.cacheReturnValue
  def exists(self):
    #io.dbg(f'self.exists: {self.key=}, {[b.path for b in self.bibs]=}')
    if self.bibs is None:
      return None
    for b in self.bibs:
      if self in b.cites():
        #io.dbg(f'exsists in {b.path}')
        return True
    return False


  def toString(self):
    fieldsStr = ''
    for k, v in self.fields.items():
      fieldsStr += f'  {k} = {v},\n'
    # strip trailing comma+newline and wrap properly
    if fieldsStr:
      fieldsStr = f',\n{fieldsStr[:-2]}\n'
    return f'@{self.section}{{{self.key}'+fieldsStr+'}'


  def makeKey(self):
    res = []
    author, title, year = [self.get(k, '') for k in 'author title year'.split()]

    # add first author's lastname
    author = re.sub(r'[^a-z0-9,\s]+', '', unidecode.unidecode(author).lower())
    if author:
      firstAuthor = author.split('and')[0]
      lastName = firstAuthor.split(',')[0].split()[-1]
      res.append(lastName.lower())

    # add title
    title = re.sub(r'[^a-z0-9\s]+', '', unidecode.unidecode(title).lower())
    if title:
      segments = [s.strip() for s in title.split() 
                    if s.strip() and s not in (
                        'in', 'of', 'for', 'by', 'the', 'and', 'a', 'on')]
      res.extend(segments[:2])

    # add year
    year = re.sub(r'[^a-z0-9\s]+', '', unidecode.unidecode(year).lower())
    if year:
      res.append(year)

    return '_'.join(res)


  def pretty(self):
    fieldsStr = ''
    if self.fields is not None:
      # strip http etc. from doi field if existing
      doi = self['doi']
      if (cfg.get('bib_repair', 'repair_doi')
          and doi is not None):
        for strip in ('http://doi.org/', 'https://doi.org/'):
          if doi.startswith(strip):
            self['doi'] = doi[len(strip):]
            break

      # replace url field if doi field exists
      doi = self['doi']
      url = self['url']
      if (cfg.get('bib_repair', 'generate_url_from_doi')
                and doi is not None
                and url is None):
        self['doi'] = doi
        self['url'] = f'https://doi.org/{doi}'

      # check if url is available if connected to the internet and url present
      url = self['url']
      if (cfg.get('bib_repair', 'verify_url_exists')
          and url is not None):
        err = isUrlValid(url)
        if err:
          if 'cloudflare' in err.lower():
            _printCloudflareWarning()
            io.dbg(f'citation {self.key} has {err}')
          else:
            io.warn(f'citation {self.key} has {err}')

      # check if doi is valid if connected to the internet and doi present
      doi = self['doi']
      if (cfg.get('bib_repair', 'verify_doi_exists')
          and doi is not None):
        err = isDoiValid(doi)
        if err:
          if 'cloudflare' in err.lower():
            _printCloudflareWarning()
            io.dbg(f'citation {self.key} has {err}')
          else:
            io.warn(f'citation {self.key} has {err}')

      # reformat pages field if exists
      pages = self['pages']
      if (cfg.get('bib_repair', 'keep_only_startpage_in_pages')
                and pages is not None):
        m = re.match(r'^(\d+)', pages)
        if m:
          self['pages'] = m.groups()[0]
      elif (cfg.get('bib_repair', 'force_double_hyphen_in_pages')
                and pages is not None):
        m = re.match(r'(\d+)[-\s]+(\d+)', pages)
        if m:
          self['pages'] = m.groups()[0]+'--'+m.groups()[1]

      # convert month field to number if given in string representation
      month = self['month']
      if (cfg.get('bib_repair', 'convert_month_to_number')
                and month is not None):
        try:
          self['month'] = int(month)
        except:
          months = 'jan feb mar apr may jun jul aug sep oct nov dec'.split()
          for i, m in enumerate(months):
            if month.lower().startswith(m):
              self['month'] = i+1
              break
          else:
            io.warn(f'found invalid month "{month}" in citation "{self.key}", '
                    f'ignoring...')

      # check if config for journal conversion is consistent
      if (cfg.get('bib_repair', 'convert_journal_to_iso4_abbr')
            and cfg.get('bib_repair', 'convert_journal_to_full_name')):
        raise RuntimeError('invalid "bib_repair" config: both',
                           '"convert_journal_to_iso4_abbr" and',
                           '"convert_journal_to_full_name" are enabled')

      # search for journal in library and ask to add it if not existing
      # replace journal field with library value
      jour = self['journal']
      if ((cfg.get('bib_repair', 'convert_journal_to_iso4_abbr')
             or cfg.get('bib_repair', 'convert_journal_to_full_name'))
                and jour is not None):
        from ..parser import journal
        abbr, name = journal.getOrAsk(jour, self.toString())
        if cfg.get('bib_repair', 'convert_journal_to_iso4_abbr'):
          res = abbr
        else:
          res = name
        if res != jour:
          io.verb('replaced journal field',
                  repr(jour),
                  'with',
                  repr(res))
        self['journal'] = res

      for k, v in self.fields.items():
        # skip ignore fields
        if k in cfg.get('bib_repair', 'remove_fields'):
          continue

        # try to reformat curly braces etc. if parse info is available
        if (self.fieldsParseInfo is not None
              and k in self.fieldsParseInfo):
          opens, closes = [self.fieldsParseInfo[k][_k]
                                      for _k in ('opens', 'closes')]
          if not opens and not closes:
            if v[0] != '{' and v[-1] != '}':
              v = '{'+v+'}'

          elif (len(opens) != len(closes)
                or opens[0] != 0
                or closes[-1] != len(v)-1):
            io.dbg(f'v={v}', f'opens={opens}', f'closes={closes}')
            io.warn('unexpected curly brace and quote structure at',
                    f'citation with key "{self.key}" in item "{k}",',
                    'failed to pretty print this citation')
          else:
            # replace outer enclosure with curly braces
            opens.pop(0)
            closes.pop(-1)
            v = '{'+v[1:-1]+'}'

            # go through other opens and closes and check why they are there
            while len(opens):
              o = opens.pop(0)
              if closes[0] <= o:
                raise RuntimeError('unexpected error')
              if opens and closes[0] >= opens[0]:
                #io.warn('curly brace/quote structure is nested too deeply in',
                #        f'citation with key "{self.key}" in item "{k}",',
                #        'failed to pretty print this citation')
                break
              c = closes.pop(0)

              # detect if only single letters or fraction of a word are proteced
              _o, _c = o, c
              while _o > 0 and (v[_o].isalpha() or v[_o] in r'\-{}'):
                _o -= 1
              while _c < len(v) and (v[_c].isalpha() or v[_c] in r'\-{}'):
                _c += 1
              orig = v[o+1:c].replace('{', '').replace('}', '')
              expanded = v[_o+1:_c].replace('{', '').replace('}', '')
              askProtect = True
              if ('\\' not in v[:_c+1]
                      and orig.strip()
                      and orig != expanded
                      and all([c not in expanded for c in r'/:_'])):
                if io.conf(f'detected protected capitalization of a fraction '
                           f'of a word in',
                           f'"{v}"',
                           f'expand "{v[o+1:c]}" -> "{expanded}"?', default=True):
                  # update value and open/close positions
                  v = v[:_o+1]+'{'+v[_o+1:o]+v[o+1:c]+v[c+1:_c]+'}'+v[_c:]
                  o = _o+1
                  c = _c-1

                  # remove other openings and closings in word range
                  while opens and opens[0] <= c:
                    opens.pop(0)
                    closes.pop(0)
                  askProtect = False

              # apply/dont apply protection
              protect = v[o+1:c]
              if (protect.strip()
                    and all([_c not in v[:c] for _c in r'\/:_&'])):
                if shouldBeProtected(v, v[o+1:c].replace('{', '').replace('}', ''),
                                     askProtect=askProtect):
                  v = v[:o]+'{'+v[o+1:c]+'}'+v[c+1:]
                else:
                  v = v[:o]+v[o+1:c]+v[c+1:]
                  opens = [o-2 for o in opens]
                  closes = [c-2 for c in closes]

        # check if always protect phrases exist and offer protection
        if cfg.get('bib_repair', 'protect_capitalization_if_unprotected'):
          for protect in cfg.get('z_bib_words_protect_capitalization'):
            protect = protect.strip()
            i = -1
            while True:
              i = v.find(protect, i+1)
              if i < 0:
                break
              if ((i-1 < 0
                      or v[i-1] in ' ')
                  and (i+len(protect) > len(v)-2
                      or v[i+len(protect)] in ' ')):
                v = v[:i]+'{'+v[i:i+len(protect)]+'}'+v[i+len(protect):]

        if cfg.get('bib_repair', 'make_all_items_lowercase'):
          k = k.lower()

        fieldsStr += f'  {k} = {v},\n'

      # strip trailing comma+newline and wrap properly
      if fieldsStr:
        fieldsStr = f',\n{fieldsStr[:-2]}\n'

    res = f'@{self.section}{{{self.key}'+fieldsStr+'}'
    #io.dbg('pretty printed citation:', res)

    return res



import os

from .. import io
from . import common


class ImgFile:
  def __init__(self, fname, paths=None):
    self.fname = fname
    self.paths = paths
    self.path = None
    if paths is None:
      self.path = os.path.normpath(fname)
    else:
      for p in list(paths)+['.']:
        _p = os.path.join(p, os.path.dirname(fname))
        if os.path.isdir(_p):
          for f in os.listdir(_p):
            if common.filenamesEqual(f, os.path.basename(fname)):
              self.path = os.path.join(p, f)
              break
          if self.path is not None:
            break


  def containsNewcommandArg(self):
    return common.containsNewcommandArg(self.fname)


  def __eq__(self, img):
    if self.path and img.path:
      return common.filenamesEqual(self.path, img.path)
    return common.filenamesEqual(os.path.basename(self.fname),
                                os.path.basename(img.fname))


  def __lt__(self, img):
    return self.path and img.path and self.path < img.path


  def __gt__(self, img):
    return self.path and img.path and self.path > img.path


  def exists(self):
    return self.path and os.path.exists(self.path)


  def __repr__(self):
    return f'img file: {self.path or self.fname}'



from .. import io

def main(args):
  authorList = [a.split() for a in ' '.join(
                  [args.separator if a == 'and' else a
                      for a in args.authors]).split(args.separator)]
  first = last = []
  if args.keep_first or args.keep_first_and_last:
    first = [authorList.pop(0)]
  if args.keep_last or args.keep_first_and_last:
    last = [authorList.pop(-1)]
  sortedAuthorList = sorted(authorList, key=lambda a: a[-1])
  authors = first+sortedAuthorList+last
  if not args.quiet:
    if any([len(a) != 2 for a in authors]):
      io.warn('there were names with more than one surname and one family name, '
              'they were treated as if all names except the last were surnames')
    if authorList == sortedAuthorList:
      io.info("list was properly sorted")
    else:
      io.info("list was _not_ properly sorted")
  if args.no_and:
    finalSep = args.separator.strip()+' '
  else:
    finalSep = ' and '
  def fmtAuthor(author, full=args.full_name):
    if full:
      return ' '.join([a.capitalize() for a in author])
    else:
      return (' '.join([a[0].upper()+'.' for a in author[:-1]])
                        +' '+author[-1].capitalize())
  io.info((args.separator.strip()+' ').join([fmtAuthor(a) for a in authors[:-1]])
            +finalSep+fmtAuthor(authors[-1]))



import os

from .. import io

def pathRelTo(file, name):
  if type(file) is str:
    path = file
  elif hasattr(file, 'toplevel') and file.toplevel:
    return pathRelTo(file.toplevel, name)
  elif hasattr(file, 'path'):
    path = file.path
  return os.path.normpath(
             os.path.join(
                 os.path.dirname(path), name))


def containsNewcommandArg(string):
  return '#' in string


def stripSuffix(n):
  res = n
  if '.' in n:
    i = n.rfind('.')
    if os.path.sep not in n[i:]:
      res = n[:i]
  return res


def filenamesEqual(n1, n2):
  return stripSuffix(n1) == stripSuffix(n2)



import sys
import os
import subprocess
import re
import shutil
import time

from .common import *
from .. import cfg
from .. import io

p = lambda d: os.path.realpath(os.path.expanduser(d))
buildFilesTrash = p(cfg.get('clean', 'trash_folder'))
baseDirectory = p('.')
ignorePaths   = [p(d) for d in cfg.get('clean', 'ignore_paths')]
requiredSuffs = cfg.get('clean', 'required_suffs')
cleanSuffs = cfg.get('clean', 'clean_suffs')


def stripBaseDir(p):
  if p.startswith(baseDirectory):
    return p[len(baseDirectory):]
  return p


def main(args):
  m = []
  for root, dirs, files in os.walk(baseDirectory, topdown=True):
    i = 0
    while i < len(dirs):
      if dirs[i].startswith('.'):
        dirs.pop(i)
      else:
        i += 1
    if (all([any(['.'+rs in f for f in files]) for rs in requiredSuffs])
        and not any([root.startswith(ip) for ip in ignorePaths])):
      m.extend([f'{root}/{f}' for f in files
                        if any(['.'+cs in f for cs in cleanSuffs])])

  if len(m):
    io.info('found build files:', *m)
    if io.conf(f'move all found build files to {buildFilesTrash}?',
               default=False):
      for path in m:
        absPath, path = path, stripBaseDir(path)
        baseFolder = time.strftime('%Y-%m-%d')
        absMoveTo = os.path.realpath(f'{buildFilesTrash}/{baseFolder}/{path}')
        io.info(f'{stripBaseDir(absPath)} -> {stripBaseDir(absMoveTo)}')
        os.makedirs(os.path.dirname(absMoveTo), exist_ok=True)
        shutil.move(absPath, absMoveTo)
      io.info('all done.')
  else:
    io.info(f'no tex build files found in {baseDirectory}')



import time
import traceback
import os

from . import io
from . import cfg


def uniquePath(path):
  i = 0
  _path = path
  while os.path.exists(path):
    p = _path.split('.')
    if len(p) > 1:
      path = '.'.join(p[:-1])+f'-{i}.'+p[-1]
    i += 1
  return path


def logExceptionsAndRaise(func):
  def wrapper(*args, **kwargs):
    try:
      return func(*args, **kwargs)
    except KeyboardInterrupt:
      raise
    except:
      io.err(traceback.format_exc(), logOnly=True)
      raise
  return wrapper


def cacheReturnValue(func):
  def wrapper(*args, **kwargs):
    self = func
    if len(args) > 0:
      try:
        setattr(args[0], '__dummy_attribute_not_used_anywhere___', None)
        self = args[0]
        #io.dbg(f'cacheReturnValue decorator detected class '
        #       f'{self.__class__.__name__}')
      except:
        pass
    if self is func:
      pass
      #io.dbg(f'cacheReturnValue decorator detected function '
      #       f'{self.__name__}')
    cacheAttr = '_'+func.__name__
    if (not hasattr(self, cacheAttr)
         or getattr(self, cacheAttr) is None):
      setattr(self, cacheAttr, func(*args, **kwargs))
    return getattr(self, cacheAttr)
  return wrapper


def replaceSuffix(path, suff):
  suff = suff.strip()
  fname = os.path.basename(path)
  if '.' in fname:
    fname = '.'.join(fname.split('.')[:-1])
  if len(suff)>0 and not suff.startswith('.'):
    suff = '.'+suff
  return os.path.join(os.path.dirname(path), fname+suff)



import os

from .. import project
from .. import cfg
from .. import io

# return if a path looks like a valid image directory according
# to the configured image directory name
def isGoodImageDirectory(p):
  return (os.path.sep+cfg.get('img_dir_name')+os.path.sep
                        in os.path.sep+p+os.path.sep)


# return prettified directory paths: img -> ./img/, ../a/b -> ../a/b/
def prettyDirectory(p):
  if p.startswith('.') or p.startswith(os.path.sep):
    res = p
  else:
    res = '.'+os.path.sep+p
  if not res.endswith(os.path.sep):
    res += os.path.sep
  return res


# detect project from either given toplevel path of current directory
# warn and offer to cancel if too many toplevel files are detected
def detectProj(args, **kwargs):
  proj = project.Project(args, **kwargs)

  if len(proj.toplevel()) > 3:
    if len(proj.toplevel()) > 20:
      io.info('detected toplevel files: ', *sorted(proj.toplevel())[:20], '...')
    else:
      io.info('detected toplevel files: ', *sorted(proj.toplevel()))
    if not io.conf(f'detected {len(proj.toplevel())} toplevel files, continue?',
                   default=False):
      return None
  return proj



import re

from .. import io
from .. import cfg


def _beginsMatch(string, query):
  q = query.lower().split()
  for s in string.lower().split():
    if q and s.startswith(q[0]):
      q.pop(0)
  return len(q) == 0


def _filtAbbr(a):
  return re.sub(r'\([^)]*\)', '', a).strip()


def _filtName(n):
  return re.sub(r'\([^)]*\)', '', n).strip()


def find(*name):
  matches = []
  for k, v in journals()[-1].items():
    k, v = _filtAbbr(k), _filtName(v)
    if (_beginsMatch(k, ' '.join(name))
        or _beginsMatch(v, ' '.join(name))):
      if [k, v] not in matches:
        matches.append([k, v])
  matches = sorted(matches, key=lambda m: m[0])
  for k, v in journals()[-1].items():
    k, v = _filtAbbr(k), _filtName(v)
    if (all([n.lower() in k.lower() for n in name])
        or all([n.lower() in v.lower() for n in name])):
      if [k, v] not in matches:
        matches.append([k, v])
  return matches


def _compareStr(s):
  return ' '.join([re.sub(r'[{}\\"\.:-_]', '', _s.strip())
                    for _s in s.split(':')[0].lower().split()])


def byAbbr(abbr):
  abbr = _compareStr(abbr)
  keys, labbrs, _, j = journals()
  if abbr in labbrs:
    k = keys[labbrs.index(abbr.lower())]
    return k, j[k]
  return None, None


def byName(name):
  name = _compareStr(name)
  keys, _, lnames, j = journals()
  if name in lnames:
    k = keys[lnames.index(name.lower())]
    return k, j[k]
  return None, None


def getOrAsk(jour=None, descr=None):
  # check if journal exists in db
  k, v = byName(jour)
  if k and v:
    return k, v

  k, v = byAbbr(jour)
  if k and v:
    return k, v

  # else show full bibinfo
  if descr:
    io.info('in bibtex entry:', descr)

  # and check if similar journals exist in db
  matches = find(jour)
  options = [m[1] for m in matches[:9]] + ['<manually enter journal>']
  
  # suggest generated abbreviation if pyiso4 is available
  try:
    import pyiso4.ltwa
    autoAbbrev = pyiso4.ltwa.Abbreviate.create()(jour)
    matches = [(autoAbbrev, jour)] + matches
    options = ['<guessed abbreviation: '+autoAbbrev+'>'] + options
  except ImportError:
    pass  

  if matches:
    i, j = io.select(f'journal {repr(jour)} not found in database,\n'
                     f'if this journal is among one of the following\n'
                     f'similar journals select it, else select to\n'
                     f'manually enter journal details', options)
    if i < len(options)-1:
      return matches[i]

  else:
    io.info(f'journal {repr(jour)} not found in database')

  while True:
    io.info('please enter full name of the journal:')
    while True:
      full = input().strip()
      if full:
        break

    addToDb = True
    k, v = byName(full)
    if k and v and io.conf('found this journal in database, use database entry?',
                           default=True):
      addToDb = False
      abbr = k
      full = v
    else:
      io.info('please enter ISO4 abbreviated name of the journal:')
      while True:
        abbr = input().strip()
        if abbr:
          break

    if io.conf(f'the correct full name for the journal',
               repr(jour),
               f'is',
               repr(full),
               f'the correct ISO4 abbreviated name is',
               repr(abbr),
               f'is this correct?'):
      break
  if addToDb:
    cfgJournals = cfg.get('z_bib_journals')
    cfgJournals[abbr] = full
    cfg.set('z_bib_journals', cfgJournals)
    journals(forceRegen=True)
  return byName(full)


def printSuspicious():
  for k, v in journals()[-1].items():
    if any([not w.isupper()
              and not w.isdigit()
              and not w.endswith('.')
              and not len(w) > 3
              and not w.lower() in ('new low ion in of'.split())
                          for w in k.split() if w]):
      io.info(f'suspicious abbreviation: "{k}", "{v}"')
    elif any([w.endswith('.') for w in v.split() if w]):
      io.info(f'suspicious full name: "{k}", "{v}"')
    elif any([c in s for c in '()' for s in (k, v)]):
      io.info(f'suspicious characters in: "{k}", "{v}"')


_JOURNALS = None
_KEYS = None
_LKEYS = None
_LNAMES = None

def journals(forceRegen=False):
  global _JOURNALS, _KEYS, _LKEYS, _LNAMES
  if _JOURNALS is None or forceRegen:
    _JOURNALS = {}
    _JOURNALS.update(JOURNALS)
    _JOURNALS.update(CAS_JOURNALS)
    _JOURNALS.update(cfg.get('z_bib_journals'))
    _JOURNALS = {_filtAbbr(k): _filtName(v) for k, v in _JOURNALS.items()}
    _KEYS = list(_JOURNALS.keys())
    _LKEYS = [_compareStr(k) for k in _KEYS]
    _LNAMES = [_compareStr(_JOURNALS[k]) for k in _KEYS]
  return _KEYS, _LKEYS, _LNAMES, _JOURNALS


# manually managed list
JOURNALS = {
  'Physica A': 'Physica A: Statistical Mechanics and its Applications',
  'Physica B Condens. Matter': 'Physica B: Condensed Matter',
  'Physica C Supercond.': 'Physica C: Superconductivity and its Applications',
  'Physica D': 'Physica D: Nonlinear Phenomena',
  'Physica E Low Dimens. Syst. Nanostruct.': 'Physica E: Low-dimensional Systems and Nanostructures',
  'AIP Conf. Proc.': 'AIP Conference Proceedings',
  'Adv. Phys.': 'Advances in Physics',
  'Am. Sci.': 'American Scientist',
  'Ann. Phys.': 'Annalen der Physik',
  'Br. J. Philos. Sci.': 'British Journal for the Philosophy of Science',
  'Commun. Math. Phys.': 'Communications in Mathematical Physics',
  'Contemp. Phys.': ' Contemporary Physics',
  'Czechoslov. J. Phys.': ' Czechoslovak Journal of Physics',
  'Z. Phys. A': '{Zeitschrift} {f\\"ur} {Physik} A Hadrons and Nuclei',
  'Z. Phys. B': '{Zeitschrift} {f\\"ur} {Physik} B Condensed Matter',
  'Z. Phys. C': '{Zeitschrift} {f\\"ur} {Physik} C Particles and Fields',
  'Z. Phys. D': '{Zeitschrift} {f\\"ur} {Physik} D Atoms, Molecules and Clusters',
  'Found. Phys.': ' Foundations of Physics',
  'IBM J. Res. Dev.': ' IBM Journal of Research and Development',
  'J. Math. Phys.': ' Journal of Mathematical Physics',
  'J. Opt. Soc. Am. A': 'Journal of the Optical Society of America A',
  'J. Opt. Soc. Am. B': ' Journal of the Optical Society of America B',
  'J. Phys. A': 'Journal of Physics A: Mathematical and General',
  'J. Phys. B': 'Journal of Physics B: Atomic, Molecular and Optical Physics',
  'J. Phys. C': 'Journal of Physics C: Solid State Physics',
  'J. Phys. D': 'Journal of Physics D: Applied Physics',
  'J. Phys. E': 'Journal of Physics E: Scientific Instruments',
  'J. Phys. F': 'Journal of Physics F: Metal Physics',
  'J. Phys. G': 'Journal of Physics G: Nuclear and Particle Physics',
  'J. Phys. Soc. Jpn.': ' Journal of the Physical Society of Japan',
  'J. Stat. Phys.': ' Journal of Statistical Physics',
  'J. Supercond. Nov. Magn.': ' Journal of Superconductivity and Novel Magnetism',
  'Nat. Rev. Mater.': ' Nature Reviews Materials',
  'Nature': 'Nature',
  'New J. Phys.': ' New Journal of Physics',
  'Opt. Act. Int. J. Opt.': 'Optica Acta: International Journal of Optics',
  'Phys. Rep.': 'Physics Reports',
  'Phys. Rev.': 'Physical Review Journals Archive',
  'Rep. Prog. Phys.': 'Reports on Progress in Physics',
  'Rev. Mod. Phys.': 'Reviews of Modern Physics',
  'Science': 'Science',
  'Z. Phys.': '{Zeitschrift} {f\\"ur} {Physik}',
  'arXiv': 'arXiv',

  # corrections of CAS database, where text after colon in abbreviated form exists:
  'Acta Crystallogr., Sect. A': 'Acta Crystallographica, Section A: Foundations and Advances',
  'Acta Crystallogr., Sect. B': 'Acta Crystallographica, Section B: Structural Science, Crystal Engineering and Materials',
  'Acta Crystallogr., Sect. C': 'Acta Crystallographica, Section C: Structural Chemistry',
  'Acta Crystallogr., Sect. D': 'Acta Crystallographica, Section D: Biological Crystallography',
  'Acta Crystallogr., Sect. E': 'Acta Crystallographica, Section E: Crystallographic Communications',
  'Acta Crystallogr., Sect. F': 'Acta Crystallographica, Section F: Structural Biology Communications',
  'Appl. Phys. A': 'Applied Physics A: Materials Science &amp; Processing',
  'Appl. Phys. B': 'Applied Physics B: Lasers and Optics',
  'Comp. Biochem. Physiol., Part A': 'Comparative Biochemistry and Physiology, Part A: Molecular &amp; Integrative Physiology',
  'Comp. Biochem. Physiol., Part B': 'Comparative Biochemistry and Physiology, Part B: Biochemistry &amp; Molecular Biology',
  'Comp. Biochem. Physiol., Part C': 'Comparative Biochemistry and Physiology, Part C: Toxicology &amp; Pharmacology',
  'Comp. Biochem. Physiol., Part D': 'Comparative Biochemistry and Physiology, Part D: Genomics &amp; Proteomics',
  'Phys. Rev. A': 'Physical Review A: Atomic, Molecular, and Optical Physics',
  'Phys. Rev. B': 'Physical Review B: Condensed Matter and Materials Physics',
  'Phys. Rev. C': 'Physical Review C: Nuclear Physics',
  'Phys. Rev. D': 'Physical Review D: Particles, Fields, Gravitation, and Cosmology',
  'Phys. Rev. E': 'Physical Review E: Statistical, Nonlinear, and Soft Matter Physics',

  # corrections of CAS database, where fuer was used instaed of {f\\"ur}:
  'Monatsh. Chem.': 'Monatshefte {f\\"ur} Chemie',
  'Neues Jahrb. Mineral., Abh.': 'Neues Jahrbuch {f\\"ur} Mineralogie, Abhandlungen',
  'Z. Anorg. Allg. Chem.': 'Zeitschrift {f\\"ur} Anorganische und Allgemeine Chemie',
  'Z. Lebensm.-Unters. Forsch. A': 'Zeitschrift {f\\"ur} Lebensmittel-Untersuchung und -Forschung A: European Food Research and Technology',
  'Z. Naturforsch., A': 'Zeitschrift {f\\"ur} Naturforschung, A: Physical Sciences',
  'Z. Naturforsch., B': 'Zeitschrift {f\\"ur} Naturforschung, B: A Journal of Chemical Sciences',
  'Z. Naturforsch., C': 'Zeitschrift {f\\"ur} Naturforschung, C: Journal of Biosciences',
  'Z. Phys. Chem.': 'Zeitschrift {f\\"ur} Physikalische Chemie',

}

# from: https://www.cas.org/support/documentation/references/corejournals
CAS_JOURNALS = {
  'ACS Appl. Mater. Interfaces': 'ACS Applied Materials &amp; Interfaces',
  'ACS Biomater. Sci. Eng.': 'ACS Biomaterials Science &amp; Engineering',
  'ACS Catal.': 'ACS Catalysis',
  'ACS Cent. Sci.': 'ACS Central Science',
  'ACS Chem. Biol.': 'ACS Chemical Biology',
  'ACS Chem. Neurosci.': 'ACS Chemical Neuroscience',
  'ACS Comb. Sci.': 'ACS Combinatorial Science',
  'ACS Infect. Dis.': 'ACS Infectious Diseases',
  'ACS Macro Lett.': 'ACS Macro Letters',
  'ACS Med. Chem. Lett.': 'ACS Medicinal Chemistry Letters',
  'ACS Sens.': 'ACS Sensors',
  'ACS Sustainable Chem. Eng.': 'ACS Sustainable Chemistry &amp; Engineering',
  'ACS Symp. Ser.': 'ACS Symposium Series',
  'ACS Synth. Biol.': 'ACS Synthetic Biology',
  'AIChE J.': 'AIChE Journal',
  'APL Mater.': 'APL Materials',
  'ARKIVOC (Gainesville, FL, U. S.)': 'ARKIVOC (Gainesville, FL, United States)',
  'Acc. Chem. Res.': 'Accounts of Chemical Research',
  'Acta Biochim. Pol.': 'Acta Biochimica Polonica',
  'Acta Chim. Slovaca': 'Acta Chimica Slovaca',
  'Acta Chim. Slov.': 'Acta Chimica Slovenica',
  'Acta Crystallogr., Sect. A: Found. Adv.': 'Acta Crystallographica, Section A: Foundations and Advances',
  'Acta Crystallogr., Sect. B: Struct. Sci., Cryst. Eng. Mater.': 'Acta Crystallographica, Section B: Structural Science, Crystal Engineering and Materials',
  'Acta Crystallogr., Sect. C: Struct. Chem.': 'Acta Crystallographica, Section C: Structural Chemistry',
  'Acta Crystallogr., Sect. D: Biol. Crystallogr.': 'Acta Crystallographica, Section D: Biological Crystallography',
  'Acta Crystallogr., Sect. E: Crystallogr. Commun.': 'Acta Crystallographica, Section E: Crystallographic Communications',
  'Acta Crystallogr., Sect. F: Struct. Biol. Commun.': 'Acta Crystallographica, Section F: Structural Biology Communications',
  'Acta Mater.': 'Acta Materialia',
  'Acta Pharm. (Zagreb, Croatia)': 'Acta Pharmaceutica (Zagreb, Croatia)',
  'Acta Pharmacol. Sin.': 'Acta Pharmacologica Sinica',
  'Acta Phys. Pol. A': 'Acta Physica Polonica A',
  'Acta Physiol.': 'Acta Physiologica',
  'Adsorpt. Sci. Technol.': 'Adsorption Science &amp; Technology',
  'Adv. Drug Delivery Rev.': 'Advanced Drug Delivery Reviews',
  'Adv. Electron. Mater.': 'Advanced Electronic Materials',
  'Adv. Energy Mater.': 'Advanced Energy Materials',
  'Adv. Funct. Mater.': 'Advanced Functional Materials',
  'Adv. Healthcare Mater.': 'Advanced Healthcare Materials',
  'Adv. Mater. (Weinheim, Ger.)': 'Advanced Materials (Weinheim, Germany)',
  'Adv. Opt. Mater.': 'Advanced Optical Materials',
  'Adv. Synth. Catal.': 'Advanced Synthesis &amp; Catalysis',
  'Adv. Appl. Ceram.': 'Advances in Applied Ceramics',
  'Adv. Chem. Phys.': 'Advances in Chemical Physics',
  'Adv. Chromatogr. (Boca Raton, FL, U. S.)': 'Advances in Chromatography (Boca Raton, FL, United States)',
  'Adv. Colloid Interface Sci.': 'Advances in Colloid and Interface Science',
  'Adv. Electrochem. Sci. Eng.': 'Advances in Electrochemical Science and Engineering',
  'Adv. Enzyme Res.': 'Advances in Enzyme Research',
  'Adv. Inorg. Chem.': 'Advances in Inorganic Chemistry',
  'Adv. Polym. Sci.': 'Advances in Polymer Science',
  'Adv. Polym. Technol.': 'Advances in Polymer Technology',
  'Adv. X-Ray Anal.': 'Advances in X-Ray Analysis',
  'Aerosol Sci. Technol.': 'Aerosol Science and Technology',
  'Am. Ceram. Soc. Bull.': 'American Ceramic Society Bulletin',
  'Am. Chem. Sci. J.': 'American Chemical Science Journal',
  'Am. J. Clin. Nutr.': 'American Journal of Clinical Nutrition',
  'Am. J. Hum. Genet.': 'American Journal of Human Genetics',
  'Am. J. Physiol.': 'American Journal of Physiology',
  'Am. J. Respir. Cell Mol. Biol.': 'American Journal of Respiratory Cell and Molecular Biology',
  'Analyst (Cambridge, U. K.)': 'Analyst (Cambridge, United Kingdom)',
  'Anal. Chim. Acta': 'Analytica Chimica Acta',
  'Anal. Biochem.': 'Analytical Biochemistry',
  'Anal. Chem. (Washington, DC, U. S.)': 'Analytical Chemistry (Washington, DC, United States)',
  'Anal. Chem. Lett.': 'Analytical Chemistry Letters',
  'Anal. Lett.': 'Analytical Letters',
  'Anal. Sci.': 'Analytical Sciences',
  'Anal. Bioanal. Chem.': 'Analytical and Bioanalytical Chemistry',
  'Angew. Chem., Int. Ed.': 'Angewandte Chemie, International Edition',
  'Ann. Pharm. Fr.': 'Annales Pharmaceutiques Francaises',
  'Ann. Chim. (Cachan, Fr.)': 'Annales de Chimie (Cachan, France)',
  'Ann. Nucl. Energy': 'Annals of Nuclear Energy',
  'Ann. Nutr. Metab.': 'Annals of Nutrition &amp; Metabolism',
  'Ann. Phys. (Amsterdam, Neth.)': 'Annals of Physics (Amsterdam, Netherlands)',
  'Ann. Plast. Surg.': 'Annals of Plastic Surgery',
  'Annu. Rev. Biochem.': 'Annual Review of Biochemistry',
  'Annu. Rev. Biomed. Eng.': 'Annual Review of Biomedical Engineering',
  'Annu. Rev. Biophys.': 'Annual Review of Biophysics',
  'Annu. Rev. Cell Dev. Biol.': 'Annual Review of Cell and Developmental Biology',
  'Annu. Rev. Earth Planet. Sci.': 'Annual Review of Earth and Planetary Sciences',
  'Annu. Rev. Entomol.': 'Annual Review of Entomology',
  'Annu. Rev. Genet.': 'Annual Review of Genetics',
  'Annu. Rev. Genomics Hum. Genet.': 'Annual Review of Genomics and Human Genetics',
  'Annu. Rev. Immunol.': 'Annual Review of Immunology',
  'Annu. Rev. Mater. Res.': 'Annual Review of Materials Research',
  'Annu. Rev. Med.': 'Annual Review of Medicine',
  'Annu. Rev. Microbiol.': 'Annual Review of Microbiology',
  'Annu. Rev. Neurosci.': 'Annual Review of Neuroscience',
  'Annu. Rev. Nucl. Part. Sci.': 'Annual Review of Nuclear and Particle Science',
  'Annu. Rev. Nutr.': 'Annual Review of Nutrition',
  'Annu. Rev. Pathol.: Mech. Dis.': 'Annual Review of Pathology: Mechanisms of Disease',
  'Annu. Rev. Pharmacol. Toxicol.': 'Annual Review of Pharmacology and Toxicology',
  'Annu. Rev. Phys. Chem.': 'Annual Review of Physical Chemistry',
  'Annu. Rev. Physiol.': 'Annual Review of Physiology',
  'Annu. Rev. Phytopathol.': 'Annual Review of Phytopathology',
  'Annu. Rev. Plant Biol.': 'Annual Review of Plant Biology',
  'Annu. Rev. Virol.': 'Annual Review of Virology',
  'Anti-Cancer Agents Med. Chem.': 'Anti-Cancer Agents in Medicinal Chemistry',
  'Anti-Infect. Agents': 'Anti-Infective Agents',
  'Anti-Inflammatory Anti-Allergy Agents Med. Chem.': 'Anti-Inflammatory &amp; Anti-Allergy Agents in Medicinal Chemistry',
  'Antibiotics (Basel, Switz.)': 'Antibiotics (Basel, Switzerland)',
  'Antibiot. Khimioter.': 'Antibiotiki i Khimioterapiya',
  'Antimicrob. Agents Chemother.': 'Antimicrobial Agents and Chemotherapy',
  'Antioxid. Redox Signaling': 'Antioxidants &amp; Redox Signaling',
  'Antiviral Res.': 'Antiviral Research',
  'Appl. Biochem. Biotechnol.': 'Applied Biochemistry and Biotechnology',
  'Appl. Biochem. Microbiol.': 'Applied Biochemistry and Microbiology',
  'Appl. Catal., A': 'Applied Catalysis, A: General',
  'Appl. Catal., B': 'Applied Catalysis, B: Environmental',
  'Appl. Clay Sci.': 'Applied Clay Science',
  'Appl. Geochem.': 'Applied Geochemistry',
  'Appl. Microbiol. Biotechnol.': 'Applied Microbiology and Biotechnology',
  'Appl. Nanosci.': 'Applied Nanoscience',
  'Appl. Organomet. Chem.': 'Applied Organometallic Chemistry',
  'Appl. Phys. A: Mater. Sci. Process.': 'Applied Physics A: Materials Science &amp; Processing',
  'Appl. Phys. B: Lasers Opt.': 'Applied Physics B: Lasers and Optics',
  'Appl. Phys. Express': 'Applied Physics Express',
  'Appl. Phys. Lett.': 'Applied Physics Letters',
  'Appl. Phys. Rev.': 'Applied Physics Reviews',
  'Appl. Radiat. Isot.': 'Applied Radiation and Isotopes',
  'Appl. Spectrosc.': 'Applied Spectroscopy',
  'Appl. Surf. Sci.': 'Applied Surface Science',
  'Appl. Environ. Microbiol.': 'Applied and Environmental Microbiology',
  'Aquat. Toxicol.': 'Aquatic Toxicology',
  'Arch. Pharm. (Weinheim, Ger.)': 'Archiv der Pharmazie (Weinheim, Germany)',
  'Arch. Biochem. Biophys.': 'Archives of Biochemistry and Biophysics',
  'Arch. Environ. Contam. Toxicol.': 'Archives of Environmental Contamination and Toxicology',
  'Arch. Insect Biochem. Physiol.': 'Archives of Insect Biochemistry and Physiology',
  'Arch. Microbiol.': 'Archives of Microbiology',
  'Arch. Pharmacal Res.': 'Archives of Pharmacal Research',
  'Arch. Physiol. Biochem.': 'Archives of Physiology and Biochemistry',
  'Arch. Toxicol.': 'Archives of Toxicology',
  'Arch. Virol.': 'Archives of Virology',
  'Arterioscler., Thromb., Vasc. Biol.': 'Arteriosclerosis, Thrombosis, and Vascular Biology',
  'Asia-Pac. J. Chem. Eng.': 'Asia-Pacific Journal of Chemical Engineering',
  'Asian J. Chem.': 'Asian Journal of Chemistry',
  'Asian J. Org. Chem.': 'Asian Journal of Organic Chemistry',
  'Assay Drug Dev. Technol.': 'Assay and Drug Development Technologies',
  'Atherosclerosis (Amsterdam, Neth.)': 'Atherosclerosis (Amsterdam, Netherlands)',
  'Atmos. Chem. Phys.': 'Atmospheric Chemistry and Physics',
  'Atmos. Environ.': 'Atmospheric Environment',
  'At. Energy (N. Y., NY, U. S.)': 'Atomic Energy (New York, NY, United States)',
  'At. Spectrosc.': 'Atomic Spectroscopy',
  'Aust. J. Chem.': 'Australian Journal of Chemistry',
  'Auton. Autacoid Pharmacol.': 'Autonomic &amp; Autacoid Pharmacology',
  'Azerb. Khim. Zh.': 'Azerbaidzhanskii Khimicheskii Zhurnal',
  'BMB Rep.': 'BMB Reports',
  'Basic Clin. Pharmacol. Toxicol.': 'Basic &amp; Clinical Pharmacology &amp; Toxicology',
  'Beilstein J. Nanotechnol.': 'Beilstein Journal of Nanotechnology',
  'Beilstein J. Org. Chem.': 'Beilstein Journal of Organic Chemistry',
  'BioChem.: Indian J.': 'BioChemistry: An Indian Journal',
  'BioChip J.': 'BioChip Journal',
  'BioInorg. React. Mech.': 'BioInorganic Reaction Mechanisms',
  'Bioact. Carbohydr. Diet. Fibre': 'Bioactive Carbohydrates and Dietary Fibre',
  'Biocatal. Biotransform.': 'Biocatalysis and Biotransformation',
  'Biochem. Genet.': 'Biochemical Genetics',
  'Biochem. J.': 'Biochemical Journal',
  'Biochem. Pharmacol. (Amsterdam, Neth.)': 'Biochemical Pharmacology (Amsterdam, Netherlands)',
  'Biochem. Soc. Trans.': 'Biochemical Society Transactions',
  'Biochem. Syst. Ecol.': 'Biochemical Systematics and Ecology',
  'Biochem. Biophys. Res. Commun.': 'Biochemical and Biophysical Research Communications',
  'Biochem. Cell Biol.': 'Biochemistry and Cell Biology',
  'Biochem. Mol. Biol. Educ.': 'Biochemistry and Molecular Biology Education',
  'Biochim. Biophys. Acta, Bioenerg.': 'Biochimica et Biophysica Acta, Bioenergetics',
  'Biochim. Biophys. Acta, Biomembr.': 'Biochimica et Biophysica Acta, Biomembranes',
  'Biochim. Biophys. Acta, Gene Regul. Mech.': 'Biochimica et Biophysica Acta, Gene Regulatory Mechanisms',
  'Biochim. Biophys. Acta, Gen. Subj.': 'Biochimica et Biophysica Acta, General Subjects',
  'Biochim. Biophys. Acta, Mol. Basis Dis.': 'Biochimica et Biophysica Acta, Molecular Basis of Disease',
  'Biochim. Biophys. Acta, Mol. Cell Res.': 'Biochimica et Biophysica Acta, Molecular Cell Research',
  'Biochim. Biophys. Acta, Mol. Cell Biol. Lipids': 'Biochimica et Biophysica Acta, Molecular and Cell Biology of Lipids',
  'Biochim. Biophys. Acta, Proteins Proteomics': 'Biochimica et Biophysica Acta, Proteins and Proteomics',
  'Biochim. Biophys. Acta, Rev. Cancer': 'Biochimica et Biophysica Acta, Reviews on Cancer',
  'Bioconjugate Chem.': 'Bioconjugate Chemistry',
  'Biofuels, Bioprod. Biorefin.': 'Biofuels, Bioproducts &amp; Biorefining',
  'Biog. Amines': 'Biogenic Amines',
  'Bioinorg. Chem. Appl.': 'Bioinorganic Chemistry and Applications',
  'Biol. Pharm. Bull.': 'Biological &amp; Pharmaceutical Bulletin',
  'Biol. Chem.': 'Biological Chemistry',
  'Biol. Trace Elem. Res.': 'Biological Trace Element Research',
  'Biol. Membr.': 'Biologicheskie Membrany',
  'Biol. Reprod.': 'Biology of Reproduction',
  'Biomacromol. Mass Spectrom.': 'Biomacromolecular Mass Spectrometry',
  'Biomarkers Med.': 'Biomarkers in Medicine',
  'Biomater. Sci.': 'Biomaterials Science',
  'Biomed. Chromatogr.': 'Biomedical Chromatography',
  'Biomed. Mater. (Bristol, U. K.)': 'Biomedical Materials (Bristol, United Kingdom)',
  'Biomed. Res.': 'Biomedical Research',
  'Biomed. Khim.': 'Biomeditsinskaya Khimiya',
  'Biomol. Concepts': 'Biomolecular Concepts',
  'Biomol. NMR Assignments': 'Biomolecular NMR Assignments',
  'Bioorg. Med. Chem.': 'Bioorganic &amp; Medicinal Chemistry',
  'Bioorg. Med. Chem. Lett.': 'Bioorganic &amp; Medicinal Chemistry Letters',
  'Bioorg. Chem.': 'Bioorganic Chemistry',
  'Biopharm. Drug Dispos.': 'Biopharmaceutics &amp; Drug Disposition',
  'Biophys. Chem.': 'Biophysical Chemistry',
  'Biophys. J.': 'Biophysical Journal',
  'Biophysics (Engl. Transl.)': 'Biophysics (English Translation)',
  'Bioprocess Biosyst. Eng.': 'Bioprocess and Biosystems Engineering',
  'Bioresour. Technol.': 'Bioresource Technology',
  'Biosci. Rep.': 'Bioscience Reports',
  'Biosci., Biotechnol., Biochem.': 'Bioscience, Biotechnology, and Biochemistry',
  'Biosens. Bioelectron.': 'Biosensors &amp; Bioelectronics',
  'Biotechnol. J.': 'Biotechnology Journal',
  'Biotechnol. Lett.': 'Biotechnology Letters',
  'Biotechnol. Prog.': 'Biotechnology Progress',
  'Biotechnol. Appl. Biochem.': 'Biotechnology and Applied Biochemistry',
  'Biotechnol. Bioeng.': 'Biotechnology and Bioengineering',
  'Biotechnol. Bioprocess Eng.': 'Biotechnology and Bioprocess Engineering',
  'Birth Defects Res., Part A': 'Birth Defects Research, Part A: Clinical and Molecular Teratology',
  'Birth Defects Res., Part B': 'Birth Defects Research, Part B: Developmental and Reproductive Toxicology',
  'Birth Defects Res., Part C': 'Birth Defects Research, Part C: Embryo Today--Reviews',
  'Bone Res.': 'Bone Research',
  'BrJAC--Braz. J. Anal. Chem.': 'BrJAC--Brazilian Journal of Analytical Chemistry',
  'Brain Res.': 'Brain Research',
  'Breast Cancer Manage.': 'Breast Cancer Management',
  'Br. J. Nutr.': 'British Journal of Nutrition',
  'Br. J. Pharmacol.': 'British Journal of Pharmacology',
  'Bull. Environ. Contam. Toxicol.': 'Bulletin of Environmental Contamination and Toxicology',
  'Bull. Mater. Sci.': 'Bulletin of Materials Science',
  'Bull. Chem. Soc. Jpn.': 'Bulletin of the Chemical Society of Japan',
  'Bull. Korean Chem. Soc.': 'Bulletin of the Korean Chemical Society',
  'Bull. Russ. Acad. Sci.: Phys.': 'Bulletin of the Russian Academy of Sciences: Physics',
  'CNS Neurol. Disord.: Drug Targets': 'CNS &amp; Neurological Disorders: Drug Targets',
  'CNS Oncol.': 'CNS Oncology',
  'CPT: Pharmacometrics Syst. Pharmacol.': 'CPT: Pharmacometrics &amp; Systems Pharmacology',
  'Calcif. Tissue Int.': 'Calcified Tissue International',
  'Can. J. Chem. Eng.': 'Canadian Journal of Chemical Engineering',
  'Can. J. Chem.': 'Canadian Journal of Chemistry',
  'Can. J. Microbiol.': 'Canadian Journal of Microbiology',
  'Can. J. Phys.': 'Canadian Journal of Physics',
  'Can. J. Physiol. Pharmacol.': 'Canadian Journal of Physiology and Pharmacology',
  'Can. Metall. Q.': 'Canadian Metallurgical Quarterly',
  'Can. Mineral.': 'Canadian Mineralogist',
  'Cancer Genomics Proteomics': 'Cancer Genomics &amp; Proteomics',
  'Cancer Lett. (N. Y., NY, U. S.)': 'Cancer Letters (New York, NY, United States)',
  'Cancer Res.': 'Cancer Research',
  'Cancer Sci.': 'Cancer Science',
  'Carbohydr. Polym.': 'Carbohydrate Polymers',
  'Carbohydr. Res.': 'Carbohydrate Research',
  'Carbon Manage.': 'Carbon Management',
  'Cardiovasc. Hematol. Agents Med. Chem.': 'Cardiovascular &amp; Hematological Agents in Medicinal Chemistry',
  'Cardiovasc. Hematol. Disord.: Drug Targets': 'Cardiovascular &amp; Hematological Disorders: Drug Targets',
  'Cardiovasc. Res.': 'Cardiovascular Research',
  'Catal. Commun.': 'Catalysis Communications',
  'Catal. Lett.': 'Catalysis Letters',
  'Catal. Rev.: Sci. Eng.': 'Catalysis Reviews: Science and Engineering',
  'Catal. Sci. Technol.': 'Catalysis Science &amp; Technology',
  'Catal. Today': 'Catalysis Today',
  'Cell (Cambridge, MA, U. S.)': 'Cell (Cambridge, MA, United States)',
  'Cell Biochem. Biophys.': 'Cell Biochemistry and Biophysics',
  'Cell Biochem. Funct.': 'Cell Biochemistry and Function',
  'Cell Host Microbe': 'Cell Host &amp; Microbe',
  'Cell Metab.': 'Cell Metabolism',
  'Cell Rep.': 'Cell Reports',
  'Cell. Immunol.': 'Cellular Immunology',
  'Cell. Microbiol.': 'Cellular Microbiology',
  'Cell. Physiol. Biochem.': 'Cellular Physiology and Biochemistry',
  'Cell. Polym.': 'Cellular Polymers',
  'Cell. Reprogram.': 'Cellular Reprogramming',
  'Cell. Mol. Life Sci.': 'Cellular and Molecular Life Sciences',
  'Cellul. Chem. Technol.': 'Cellulose Chemistry and Technology',
  'Cem. Concr. Compos.': 'Cement &amp; Concrete Composites',
  'Cem. Concr. Res.': 'Cement and Concrete Research',
  'Cent. Nerv. Syst. Agents Med. Chem.': 'Central Nervous System Agents in Medicinal Chemistry',
  'Ceram. Int.': 'Ceramics International',
  'Ceram.-Silik.': 'Ceramics-Silikaty',
  'Cereal Chem.': 'Cereal Chemistry',
  'Ceska Slov. Farm.': 'Ceska a Slovenska Farmacie',
  'Chem. Sin.': 'Chemica Sinica',
  'Chem. Pharm. Bull.': 'Chemical &amp; Pharmaceutical Bulletin',
  'Chem. Biol. Drug Des.': 'Chemical Biology &amp; Drug Design',
  'Chem. Commun. (Cambridge, U. K.)': 'Chemical Communications (Cambridge, United Kingdom)',
  'Chem. Eng. Technol.': 'Chemical Engineering &amp; Technology',
  'Chem. Eng. Commun.': 'Chemical Engineering Communications',
  'Chem. Eng. J. (Amsterdam, Neth.)': 'Chemical Engineering Journal (Amsterdam, Netherlands)',
  'Chem. Eng. Res. Des.': 'Chemical Engineering Research and Design',
  'Chem. Eng. Sci.': 'Chemical Engineering Research and Design',
  'Chem. Eng. Process.': 'Chemical Engineering and Processing',
  'Chem. Geol.': 'Chemical Geology',
  'Chem. Pap.': 'Chemical Papers',
  'Chem. Phys.': 'Chemical Physics',
  'Chem. Phys. Lett.': 'Chemical Physics Letters',
  'Chem. Rec.': 'Chemical Record',
  'Chem. Res. Chin. Univ.': 'Chemical Research in Chinese Universities',
  'Chem. Res. Toxicol.': 'Chemical Research in Toxicology',
  'Chem. Rev. (Washington, DC, U. S.)': 'Chemical Reviews (Washington, DC, United States)',
  'Chem. Sci.': 'Chemical Science',
  'Chem. Sci. Trans.': 'Chemical Science Transactions',
  'Chem. Speciation Bioavailability': 'Chemical Speciation and Bioavailability',
  'Chem. Biochem. Eng. Q.': 'Chemical and Biochemical Engineering Quarterly',
  'Chem. Listy': 'Chemicke Listy',
  'Chem.-Biol. Interact.': 'Chemico-Biological Interactions',
  'Chem. Ing. Tech.': 'Chemie Ingenieur Technik',
  'Chem. Erde': 'Chemie der Erde',
  'Chem. Technol. (Kaunas, Lith.)': 'Chemine Technologija (Kaunas, Lithuania)',
  'Chem. Biodiversity': 'Chemistry &amp; Biodiversity',
  'Chem. Biol. (Oxford, U. K.)': 'Chemistry &amp; Biology (Oxford, United Kingdom)',
  'Chem. Biol. Interface': 'Chemistry &amp; Biology Interface',
  'Chem. Ind. (Chichester, U. K.)': 'Chemistry &amp; Industry (Chichester, United Kingdom)',
  'Chem. J. Mold.': 'Chemistry Journal of Moldova',
  'Chem. Lett.': 'Chemistry Letters',
  'Chem. Phys. Lipids': 'Chemistry and Physics of Lipids',
  'Chem. Technol. Fuels Oils': 'Chemistry and Technology of Fuels and Oils',
  'Chem. Heterocycl. Compd. (N. Y., NY, U. S.)': 'Chemistry of Heterocyclic Compounds (New York, NY, United States)',
  'Chem. Mater.': 'Chemistry of Materials',
  'Chem. Nat. Compd.': 'Chemistry of Natural Compounds',
  'Chemom. Intell. Lab. Syst.': 'Chemometrics and Intelligent Laboratory Systems',
  'Chemotherapy (Basel, Switz.)': 'Chemotherapy (Basel, Switzerland)',
  'Chim. Ind. (Milan, Italy)': "Chimica e l'Industria (Milan, Italy)",
  'Chin. Chem. Lett.': 'Chinese Chemical Letters',
  'Chin. J. Chem. Eng.': 'Chinese Journal of Chemical Engineering',
  'Chin. J. Chem. Phys.': 'Chinese Journal of Chemical Physics',
  'Chin. J. Chem.': 'Chinese Journal of Chemistry',
  'Chin. J. Geochem.': 'Chinese Journal of Geochemistry',
  'Chin. J. Polym. Sci.': 'Chinese Journal of Polymer Science',
  'Chin. J. Struct. Chem.': 'Chinese Journal of Structural Chemistry',
  'Chin. Phys. C': 'Chinese Physics C',
  'Chromatogr. Res. Int.': 'Chromatography Research International',
  'Chromosome Res.': 'Chromosome Research',
  'Circ. Res.': 'Circulation Research',
  'Clay Miner.': 'Clay Minerals',
  'Clays Clay Miner.': 'Clays and Clay Minerals',
  'Clin. Chim. Acta': 'Clinica Chimica Acta',
  'Clin. Chem. (Washington, DC, U. S.)': 'Clinical Chemistry (Washington, DC, United States)',
  'Clin. Chem. Lab. Med.': 'Clinical Chemistry and Laboratory Medicine',
  'Clin. Immunol. (Amsterdam, Neth.)': 'Clinical Immunology (Amsterdam, Netherlands)',
  'Clin. Proteomics': 'Clinical Proteomics',
  'Clin. Exp. Pharmacol. Physiol.': 'Clinical and Experimental Pharmacology and Physiology',
  'Clin. Vaccine Immunol.': 'Clinical and Vaccine Immunology',
  'Colloid J.': 'Colloid Journal',
  'Colloid Polym. Sci.': 'Colloid and Polymer Science',
  'Colloids Surf., A': 'Colloids and Surfaces, A: Physicochemical and Engineering Aspects',
  'Colloids Surf., B': 'Colloids and Surfaces, B: Biointerfaces',
  'Color. Technol.': 'Coloration Technology',
  'Comb. Chem. High Throughput Screening': 'Combinatorial Chemistry &amp; High Throughput Screening',
  'Combust. Sci. Technol.': 'Combustion Science and Technology',
  'Combust. Flame': 'Combustion and Flame',
  'Comments Inorg. Chem.': 'Comments on Inorganic Chemistry',
  'Commun. Soil Sci. Plant Anal.': 'Communications in Soil Science and Plant Analysis',
  'Comp. Biochem. Physiol., Part A: Mol. Integr. Physiol.': 'Comparative Biochemistry and Physiology, Part A: Molecular &amp; Integrative Physiology',
  'Comp. Biochem. Physiol., Part B: Biochem. Mol. Biol.': 'Comparative Biochemistry and Physiology, Part B: Biochemistry &amp; Molecular Biology',
  'Comp. Biochem. Physiol., Part C: Toxicol. Pharmacol.': 'Comparative Biochemistry and Physiology, Part C: Toxicology &amp; Pharmacology',
  'Comp. Biochem. Physiol., Part D: Genomics Proteomics': 'Comparative Biochemistry and Physiology, Part D: Genomics &amp; Proteomics',
  'Comp. Med.': 'Comparative Medicine',
  'Complex Met.': 'Complex Metals',
  'C. R. Chim.': 'Comptes Rendus Chimie',
  'C. R. Phys.': 'Comptes Rendus Physique',
  'Comput. Theor. Chem.': 'Computational &amp; Theoretical Chemistry',
  'Comput. Biol. Chem.': 'Computational Biology and Chemistry',
  'Comput. Phys. Commun.': 'Computer Physics Communications',
  'Comput. Chem. Eng.': 'Computers &amp; Chemical Engineering',
  'Contrast Media Mol. Imaging': 'Contrast Media &amp; Molecular Imaging',
  'Environ. Technol.': 'Environmental Technology',
  'Environ. Toxicol.': 'Environmental Toxicology',
  'Environ. Toxicol. Chem.': 'Environmental Toxicology and Chemistry',
  'Expert Rev. Vaccines': 'Expert Review of Vaccines',
  'FASEB J.': 'FASEB Journal',
  'FEBS J.': 'FEBS Journal',
  'FEBS Lett.': 'FEBS Letters',
  'Facial Plast. Surg.': 'Facial Plastic Surgery',
  'Faraday Discuss.': 'Faraday Discussions',
  'Ferroelectr., Lett. Sect.': 'Ferroelectrics, Letters Section',
  'Fibre Chem.': 'Fibre Chemistry',
  'Fire Mater.': 'Fire and Materials',
  'Fish Physiol. Biochem.': 'Fish Physiology and Biochemistry',
  'Fusion Eng. Des.': 'Fusion Engineering and Design',
  'Gene Ther.': 'Gene Therapy',
  'Gen. Comp. Endocrinol.': 'General and Comparative Endocrinology',
  'Genes Dev.': 'Genes &amp; Development',
  'Genes Cells': 'Genes to Cells',
  'Genes, Chromosomes Cancer': 'Genes, Chromosomes &amp; Cancer',
  'Hum. Gene Ther.': 'Human Gene Therapy',
  'Mol. Med. (Manhasset, NY, U. S.)': 'Molecular Medicine (Manhasset, NY, United States)',
  'Mol. Med. Rep.': 'Molecular Medicine Reports',
  'Mol. Membr. Biol.': 'Molecular Membrane Biology',
  'Mol. Metab.': 'Molecular Metabolism',
  'Mol. Microbiol.': 'Molecular Microbiology',
  'Mol. Nutr. Food Res.': 'Molecular Nutrition &amp; Food Research',
  'Mol. Pharmaceutics': 'Molecular Pharmaceutics',
  'Mol. Pharmacol.': 'Molecular Pharmacology',
  'Mol. Phys.': 'Molecular Physics',
  'Mol. Plant': 'Molecular Plant',
  'Mol. Plant-Microbe Interact.': 'Molecular Plant-Microbe Interactions',
  'Mol. Reprod. Dev.': 'Molecular Reproduction &amp; Development',
  'Mol. Simul.': 'Molecular Simulation',
  'Mol. Ther.': 'Molecular Therapy',
  'Mol. Ther.--Methods Clin. Dev.': 'Molecular Therapy--Methods &amp; Clinical Development',
  'Mol. Ther.--Nucleic Acids': 'Molecular Therapy--Nucleic Acids',
  'Mol. Cell. Biochem.': 'Molecular and Cellular Biochemistry',
  'Mol. Cell. Biol.': 'Molecular and Cellular Biology',
  'Mol. Cell. Endocrinol.': 'Molecular and Cellular Endocrinology',
  'Mol. Cell. Neurosci.': 'Molecular and Cellular Neuroscience',
  'Mol. Cell. Probes': 'Molecular and Cellular Probes',
  'Mol. Cells': 'Molecules and Cells',
  'Monatsh. Chem.': 'Monatshefte fuer Chemie',
  'Mucosal Immunol.': 'Mucosal Immunology',
  'Mutat. Res., Fundam. Mol. Mech. Mutagen.': 'Mutation Research, Fundamental and Molecular Mechanisms of Mutagenesis',
  'Mutat. Res., Genet. Toxicol. Environ. Mutagen.': 'Mutation Research, Genetic Toxicology and Environmental Mutagenesis',
  'Mutat. Res., Rev. Mutat. Res.': 'Mutation Research, Reviews in Mutation Research',
  'Nano Lett.': 'Nano Letters',
  'Nano Sci. Nano Technol.: Indian J.': 'Nano Science and Nano Technology: An Indian Journal',
  'Nanomedicine (London, U. K.)': 'Nanomedicine (London, United Kingdom)',
  'Nanomedicine (N. Y., NY, U. S.)': 'Nanomedicine (New York, NY, United States)',
  'Nat. Prod. Rep.': 'Natural Product Reports',
  'Nat. Prod. Res.': 'Natural Product Research',
  'Nat. Prod. Sci.': 'Natural Product Sciences',
  'Nat. Prod. J.': 'Natural Products Journal',
  'Nature (London, U. K.)': 'Nature (London, United Kingdom)',
  'Nat. Biotechnol.': 'Nature Biotechnology',
  'Nat. Cell Biol.': 'Nature Cell Biology',
  'Nat. Chem. Biol.': 'Nature Chemical Biology',
  'Nat. Chem.': 'Nature Chemistry',
  'Nat. Commun.': 'Nature Communications',
  'Nat. Genet.': 'Nature Genetics',
  'Nat. Geosci.': 'Nature Geoscience',
  'Nat. Immunol.': 'Nature Immunology',
  'Nat. Mater.': 'Nature Materials',
  'Nat. Med. (N. Y., NY, U. S.)': 'Nature Medicine (New York, NY, United States)',
  'Nat. Methods': 'Nature Methods',
  'Nat. Nanotechnol.': 'Nature Nanotechnology',
  'Nat. Neurosci.': 'Nature Neuroscience',
  'Nat. Photonics': 'Nature Photonics',
  'Nat. Phys.': 'Nature Physics',
  'Nat. Protoc.': 'Nature Protocols',
  'Nat. Rev. Cancer': 'Nature Reviews Cancer',
  'Nat. Rev. Drug Discovery': 'Nature Reviews Drug Discovery',
  'Nat. Rev. Genet.': 'Nature Reviews Genetics',
  'Nat. Rev. Immunol.': 'Nature Reviews Immunology',
  'Nat. Rev. Microbiol.': 'Nature Reviews Microbiology',
  'Nat. Rev. Mol. Cell Biol.': 'Nature Reviews Molecular Cell Biology',
  'Nat. Rev. Neurosci.': 'Nature Reviews Neuroscience',
  'Nat. Struct. Mol. Biol.': 'Nature Structural &amp; Molecular Biology',
  "Naunyn-Schmiedeberg's Arch. Pharmacol.": "Naunyn-Schmiedeberg's Archives of Pharmacology",
  'Neues Jahrb. Mineral., Abh.': 'Neues Jahrbuch fuer Mineralogie, Abhandlungen',
  'NeuroMol. Med.': 'NeuroMolecular Medicine',
  'Neurochem. Res.': 'Neurochemical Research',
  'Neurochem. Int.': 'Neurochemistry International',
  'Neuropeptides (Oxford, U. K.)': 'Neuropeptides (Oxford, United Kingdom)',
  'Neuroscience (Amsterdam, Neth.)': 'Neuroscience (Amsterdam, Netherlands)',
  'Neurosci. Lett.': 'Neuroscience Letters',
  'Neurotoxicol. Teratol.': 'Neurotoxicology and Teratology',
  'New Biotechnol.': 'New Biotechnology',
  'N. Engl. J. Med.': 'New England Journal of Medicine',
  'New J. Chem.': 'New Journal of Chemistry',
  'Nord. Pulp Pap. Res. J.': 'Nordic Pulp &amp; Paper Research Journal',
  'Nucl. Eng. Des.': 'Nuclear Engineering and Design',
  'Nucl. Fusion': 'Nuclear Fusion',
  'Nucl. Instrum. Methods Phys. Res., Sect. A': 'Nuclear Instruments &amp; Methods in Physics Research, Section A: Accelerators, Spectrometers, Detectors, and Associated Equipment',
  'Nucl. Instrum. Methods Phys. Res., Sect. B': 'Nuclear Instruments &amp; Methods in Physics Research, Section B: Beam Interactions with Materials and Atoms',
  'Nucl. Med. Biol.': 'Nuclear Medicine and Biology',
  'Nucl. Phys. A': 'Nuclear Physics A',
  'Nucl. Phys. B': 'Nuclear Physics B',
  'Nucleic Acid Ther.': 'Nucleic Acid Therapeutics',
  'Nucleic Acids Res.': 'Nucleic Acids Research',
  'Nucleosides, Nucleotides Nucleic Acids': 'Nucleosides, Nucleotides &amp; Nucleic Acids',
  'Nutr. Res. (N. Y., NY, U. S.)': 'Nutrition Research (New York, NY, United States)',
  'Open J. Inorg. Non-Met. Mater.': 'Open Journal of Inorganic Non-Metallic Materials',
  'Open J. Med. Chem.': 'Open Journal of Medicinal Chemistry',
  'Open J. Org. Polym. Mater.': 'Open Journal of Organic Polymer Materials',
  'Open J. Polym. Chem.': 'Open Journal of Polymer Chemistry',
  'Open J. Synth. Theory Appl.': 'Open Journal of Synthesis Theory and Applications',
  'Opt. Commun.': 'Optics Communications',
  'Opt. Spectrosc.': 'Optics and Spectroscopy',
  'Org. Biomol. Chem.': 'Organic &amp; Biomolecular Chemistry',
  'Org. Chem. Front.': 'Organic Chemistry Frontiers',
  'Org. Chem. Int.': 'Organic Chemistry International',
  'Org. Electron.': 'Organic Electronics',
  'Org. Geochem.': 'Organic Geochemistry',
  'Org. Lett.': 'Organic Letters',
  'Org. Prep. Proced. Int.': 'Organic Preparations and Procedures International',
  'Org. Process Res. Dev.': 'Organic Process Research &amp; Development',
  'Org. Synth.': 'Organic Syntheses',
  'Org. Med. Chem. Lett.': 'Organic and Medicinal Chemistry Letters',
  'Orient. J. Chem.': 'Oriental Journal of Chemistry',
  'Oxid. Commun.': 'Oxidation Communications',
  'Oxid. Met.': 'Oxidation of Metals',
  'Ozone: Sci. Eng.': 'Ozone: Science &amp; Engineering',
  'PLoS Biol.': 'PLoS Biology',
  'PLoS Comput. Biol.': 'PLoS Computational Biology',
  'PLoS Genet.': 'PLoS Genetics',
  'PLoS Pathog.': 'PLoS Pathogens',
  'PMSE Prepr.': 'PMSE Preprints',
  'Part. Sci. Technol.': 'Particulate Science and Technology',
  'Peptides (N. Y., NY, U. S.)': 'Peptides (New York, NY, United States)',
  'Pers. Med.': 'Personalized Medicine',
  'Pest Manage. Sci.': 'Pest Management Science',
  'Pestic. Biochem. Physiol.': 'Pesticide Biochemistry and Physiology',
  'Pet. Chem.': 'Petroleum Chemistry',
  'Pet. Sci. Technol.': 'Petroleum Science and Technology',
  'Pfluegers Arch.': 'Pfluegers Archiv',
  'Pharma Chem.': 'Pharma Chemica',
  'Pharm. Chem. J.': 'Pharmaceutical Chemistry Journal',
  'Pharm. Dev. Technol.': 'Pharmaceutical Development and Technology',
  'Pharm. Nanotechnol.': 'Pharmaceutical Nanotechnology',
  'Pharm. Pat. Anal.': 'Pharmaceutical Patent Analyst',
  'Pharm. Res.': 'Pharmaceutical Research',
  'Pharm. Sin.': 'Pharmacia Sinica',
  'Pharmacogenet. Genomics': 'Pharmacogenetics and Genomics',
  'Pharmacogenomics J.': 'Pharmacogenomics Journal',
  'Pharmacol. Ther.': 'Pharmacology &amp; Therapeutics',
  'Pharmacol., Biochem. Behav.': 'Pharmacology, Biochemistry and Behavior',
  'Philos. Mag.': 'Philosophical Magazine',
  'Philos. Mag. Lett.': 'Philosophical Magazine Letters',
  'Phosphorus, Sulfur Silicon Relat. Elem.': 'Phosphorus, Sulfur and Silicon and the Related Elements',
  'Photochem. Photobiol. Sci.': 'Photochemical &amp; Photobiological Sciences',
  'Photochem. Photobiol.': 'Photochemistry and Photobiology',
  'Photosynth. Res.': 'Photosynthesis Research',
  'Phys. B (Amsterdam, Neth.)': 'Physica B: Condensed Matter (Amsterdam, Netherlands)',
  'Phys. C (Amsterdam, Neth.)': 'Physica C: Superconductivity and Its Applications (Amsterdam, Netherlands)',
  'Phys. Scr.': 'Physica Scripta',
  'Phys. Scr., T': 'Physica Scripta, T',
  'Phys. Status Solidi A': 'Physica Status Solidi A: Applications and Materials Science',
  'Phys. Status Solidi B': 'Physica Status Solidi B: Basic Solid State Physics',
  'Phys. Status Solidi C': 'Physica Status Solidi C: Current Topics in Solid State Physics',
  'Phys. Status Solidi RRL': 'Physica Status Solidi RRL: Rapid Research Letters',
  'Phys. Chem. Chem. Phys.': 'Physical Chemistry Chemical Physics',
  'Phys. Rev. Appl.': 'Physical Review Applied',
  'Phys. Rev. Lett.': 'Physical Review Letters',
  'Phys. Rev. X': 'Physical Review X',
  'Phys. Lett. A': 'Physics Letters A',
  'Phys. Lett. B': 'Physics Letters B',
  'Phys. Chem. Liq.': 'Physics and Chemistry of Liquids',
  'Phys. Chem. Miner.': 'Physics and Chemistry of Minerals',
  'Phys. At. Nucl.': 'Physics of Atomic Nuclei',
  'Phys. Fluids': 'Physics of Fluids',
  'Phys. Plasmas': 'Physics of Plasmas',
  'Phys. Solid State': 'Physics of the Solid State',
  'Physiol. Plant.': 'Physiologia Plantarum',
  'Physiol. Behav.': 'Physiology &amp; Behavior',
  'Phytochem. Anal.': 'Phytochemical Analysis',
  'Phytochem. Rev.': 'Phytochemistry Reviews',
  'Plant Cell Rep.': 'Plant Cell Reports',
  'Plant Growth Regul.': 'Plant Growth Regulation',
  'Plant J.': 'Plant Journal',
  'Plant Mol. Biol.': 'Plant Molecular Biology',
  'Plant Physiol.': 'Plant Physiology',
  'Plant Physiol. Biochem. (Issy-les-Moulineaux, Fr.)': 'Plant Physiology and Biochemistry (Issy-les-Moulineaux, France)',
  'Plant Sci. (Shannon, Irel.)': 'Plant Science (Shannon, Ireland)',
  'Plant Cell Physiol.': 'Plant and Cell Physiology',
  'Plant Soil': 'Plant and Soil',
  'Planta Med.': 'Planta Medica',
  'Plasma Chem. Plasma Process.': 'Plasma Chemistry and Plasma Processing',
  'Plasma Processes Polym.': 'Plasma Processes and Polymers',
  'Plast. Reconstr. Surg.': 'Plastic and Reconstructive Surgery',
  'Plast. Massy': 'Plasticheskie Massy',
  'Plast., Rubber Compos.': 'Plastics, Rubber and Composites',
  'Pol. J. Chem. Technol.': 'Polish Journal of Chemical Technology',
  'Polycyclic Aromat. Compd.': 'Polycyclic Aromatic Compounds',
  'Polym. Bull. (Heidelberg, Ger.)': 'Polymer Bulletin (Heidelberg, Germany)',
  'Polym. Chem.': 'Polymer Chemistry',
  'Polym. Compos.': 'Polymer Composites',
  'Polym. Degrad. Stab.': 'Polymer Degradation and Stability',
  'Polym. Eng. Sci.': 'Polymer Engineering &amp; Science',
  'Polym. Int.': 'Polymer International',
  'Polym. J. (Tokyo, Jpn.)': 'Polymer Journal (Tokyo, Japan)',
  'Polym. Prepr. (Am. Chem. Soc., Div. Polym. Chem.)': 'Polymer Preprints (American Chemical Society, Division of Polymer Chemistry)',
  'Polym. Rev. (Philadelphia, PA, U. S.)': 'Polymer Reviews (Philadelphia, PA, United States)',
  'Polym. Sci., Ser. A': 'Polymer Science, Series A',
  'Polym. Sci., Ser. B': 'Polymer Science, Series B',
  'Polym. Sci., Ser. C': 'Polymer Science, Series C',
  'Polym. Sci., Ser. D': 'Polymer Science, Series D',
  'Polym. Test.': 'Polymer Testing',
  'Polym.-Plast. Technol. Eng.': 'Polymer-Plastics Technology and Engineering',
  'Polym. Polym. Compos.': 'Polymers &amp; Polymer Composites',
  'Polymers (Basel, Switz.)': 'Polymers (Basel, Switzerland)',
  'Polym. Adv. Technol.': 'Polymers for Advanced Technologies',
  'Powder Diffr.': 'Powder Diffraction',
  'Powder Metall.': 'Powder Metallurgy',
  'Powder Metall. Met. Ceram.': 'Powder Metallurgy and Metal Ceramics',
  'Powder Technol.': 'Powder Technology',
  'Pract. Spectrosc.': 'Practical Spectroscopy',
  'Precambrian Res.': 'Precambrian Research',
  'Prep. Biochem. Biotechnol.': 'Preparative Biochemistry &amp; Biotechnology',
  'Procedia Mater. Sci.': 'Procedia Materials Science',
  'Proc. Natl. Acad. Sci. U. S. A.': 'Proceedings of the National Academy of Sciences of the United States of America',
  'Process Biochem. (Oxford, U. K.)': 'Process Biochemistry (Oxford, United Kingdom)',
  'Process Saf. Environ. Prot.': 'Process Safety and Environmental Protection',
  'Prog. Colloid Polym. Sci.': 'Progress in Colloid and Polymer Science',
  'Prog. Cryst. Growth Charact. Mater.': 'Progress in Crystal Growth and Characterization of Materials',
  'Prog. Inorg. Chem.': 'Progress in Inorganic Chemistry',
  'Prog. Nucl. Magn. Reson. Spectrosc.': 'Progress in Nuclear Magnetic Resonance Spectroscopy',
  'Prog. Org. Coat.': 'Progress in Organic Coatings',
  'Prog. Polym. Sci.': 'Progress in Polymer Science',
  'Prog. React. Kinet. Mech.': 'Progress in Reaction Kinetics and Mechanism',
  'Prog. Solid State Chem.': 'Progress in Solid State Chemistry',
  'Prog. Surf. Sci.': 'Progress in Surface Science',
  'Propellants, Explos., Pyrotech.': 'Propellants, Explosives, Pyrotechnics',
  'Prostaglandins Other Lipid Mediators': 'Prostaglandins and Other Lipid Mediators',
  'Prostaglandins, Leukotrienes Essent. Fatty Acids': 'Prostaglandins, Leukotrienes and Essential Fatty Acids',
  'Prot. Met. Phys. Chem. Surf.': 'Protection of Metals and Physical Chemistry of Surfaces',
  'Protein Pept. Lett.': 'Protein &amp; Peptide Letters',
  'Protein Eng., Des. Sel.': 'Protein Engineering, Design &amp; Selection',
  'Protein Expression Purif.': 'Protein Expression and Purification',
  'Protein J.': 'Protein Journal',
  'Protein Sci.': 'Protein Science',
  'Proteins: Struct., Funct., Bioinf.': 'Proteins: Structure, Function, and Bioinformatics',
  'Proteomics: Clin. Appl.': 'Proteomics: Clinical Applications',
  'Przem. Chem.': 'Przemysl Chemiczny',
  'Psychopharmacology (Heidelberg, Ger.)': 'Psychopharmacology (Heidelberg, Germany)',
  'Pulm. Pharmacol. Ther.': 'Pulmonary Pharmacology &amp; Therapeutics',
  'Pure Appl. Chem.': 'Pure and Applied Chemistry',
  'Quantum Electron.': 'Quantum Electronics',
  'Quim. Nova': 'Quimica Nova',
  'RSC Adv.': 'RSC Advances',
  'Radiat. Eff. Defects Solids': 'Radiation Effects and Defects in Solids',
  'Radiat. Meas.': 'Radiation Measurements',
  'Radiat. Phys. Chem.': 'Radiation Physics and Chemistry',
  'Radiat. Res.': 'Radiation Research',
  'Radiochemistry (Moscow, Russ. Fed.)': 'Radiochemistry (Moscow, Russian Federation)',
  'Radiochim. Acta': 'Radiochimica Acta',
  'Rapid Commun. Mass Spectrom.': 'Rapid Communications in Mass Spectrometry',
  'Rare Met. (Beijing, China)': 'Rare Metals (Beijing, China)',
  'React. Kinet., Mech. Catal.': 'Reaction Kinetics, Mechanisms and Catalysis',
  'React. Funct. Polym.': 'Reactive &amp; Functional Polymers',
  'Recent Adv. Cardiovasc. Drug Discovery': 'Recent Advances in Cardiovascular Drug Discovery',
  'Recent Adv. DNA Gene Sequences': 'Recent Advances in DNA &amp; Gene Sequences',
  'Recent Pat. Anti-Cancer Drug Discovery': 'Recent Patents on Anti-Cancer Drug Discovery',
  'Recent Pat. Anti-Infect. Drug Discovery': 'Recent Patents on Anti-Infective Drug Discovery',
  'Recent Pat. CNS Drug Discovery': 'Recent Patents on CNS Drug Discovery',
  'Recent Pat. Drug Delivery Formulation': 'Recent Patents on Drug Delivery &amp; Formulation',
  'Recent Pat. Endocr., Metab. Immune Drug Discovery': 'Recent Patents on Endocrine, Metabolic &amp; Immune Drug Discovery',
  'Recent Pat. Inflammation Allergy Drug Discovery': 'Recent Patents on Inflammation &amp; Allergy Drug Discovery',
  'Recent Pat. Nanotechnol.': 'Recent Patents on Nanotechnology',
  'Redox Biol.': 'Redox Biology',
  'Regener. Med.': 'Regenerative Medicine',
  'Reproduction (Bristol, U. K.)': 'Reproduction (Bristol, United Kingdom)',
  'Reprod. Toxicol.': 'Reproductive Toxicology',
  'Res. Chem. Intermed.': 'Research on Chemical Intermediates',
  'Rev. Sci. Instrum.': 'Review of Scientific Instruments',
  'Rev. Anal. Chem.': 'Reviews in Analytical Chemistry',
  'Rev. Inorg. Chem.': 'Reviews in Inorganic Chemistry',
  'Rev. Mex. Ing. Quim.': 'Revista Mexicana de Ingenieria Quimica',
  'Rev. Peru. Quim. Ing. Quim.': 'Revista Peruana de Quimica e Ingenieria Quimica',
  'Rev. Rom. Mater.': 'Revista Romana de Materiale',
  'Rev. Chim. (Bucharest, Rom.)': 'Revista de Chimie (Bucharest, Romania)',
  'Rev. Soc. Quim. Peru': 'Revista de la Sociedad Quimica del Peru',
  'Rheol. Acta': 'Rheologica Acta',
  'Rubber Chem. Technol.': 'Rubber Chemistry and Technology',
  'Russ. Chem. Bull.': 'Russian Chemical Bulletin',
  'Russ. Chem. Rev.': 'Russian Chemical Reviews',
  'Russ. J. Appl. Chem.': 'Russian Journal of Applied Chemistry',
  'Russ. J. Bioorg. Chem.': 'Russian Journal of Bioorganic Chemistry',
  'Russ. J. Coord. Chem.': 'Russian Journal of Coordination Chemistry',
  'Russ. J. Electrochem.': 'Russian Journal of Electrochemistry',
  'Russ. J. Gen. Chem.': 'Russian Journal of General Chemistry',
  'Russ. J. Genet.': 'Russian Journal of Genetics',
  'Russ. J. Inorg. Chem.': 'Russian Journal of Inorganic Chemistry',
  'Russ. J. Org. Chem.': 'Russian Journal of Organic Chemistry',
  'Russ. J. Phys. Chem. A': 'Russian Journal of Physical Chemistry A',
  'Russ. J. Phys. Chem. B': 'Russian Journal of Physical Chemistry B',
  'Russ. J. Plant Physiol.': 'Russian Journal of Plant Physiology',
  'SAR QSAR Environ. Res.': 'SAR and QSAR in Environmental Research',
  'SPE Prod. Oper.': 'SPE Production &amp; Operations',
  'SPE Reservoir Eval. Eng.': 'SPE Reservoir Evaluation &amp; Engineering',
  'Science (Washington, DC, U. S.)': 'Science (Washington, DC, United States)',
  'Sci. China: Chem.': 'Science China: Chemistry',
  'Sci. Lett.': 'Science Letters',
  'Sci. Synth., Knowl. Updates': 'Science of Synthesis, Knowledge Updates',
  'Sci. Total Environ.': 'Science of the Total Environment',
  'Sci. Pharm.': 'Scientia Pharmaceutica',
  'Sci. Rep.': 'Scientific Reports',
  'Scr. Mater.': 'Scripta Materialia',
  'Semicond. Sci. Technol.': 'Semiconductor Science and Technology',
  'Sens. Actuators, A': 'Sensors and Actuators, A: Physical',
  'Sens. Actuators, B': 'Sensors and Actuators, B: Chemical',
  'Sep. Sci. Technol. (Philadelphia, PA, U. S.)': 'Separation Science and Technology (Philadelphia, PA, United States)',
  'Sep. Purif. Technol.': 'Separation and Purification Technology',
  'Soil Biol. Biochem.': 'Soil Biology &amp; Biochemistry',
  'Soil Sci.': 'Soil Science',
  'Soil Sci. Soc. Am. J.': 'Soil Science Society of America Journal',
  'Sol. Energy Mater. Sol. Cells': 'Solar Energy Materials &amp; Solar Cells',
  'Solid Fuel Chem.': 'Solid Fuel Chemistry',
  'Solid State Commun.': 'Solid State Communications',
  'Solid State Nucl. Magn. Reson.': 'Solid State Nuclear Magnetic Resonance',
  'Solid State Sci.': 'Solid State Sciences',
  'Solid-State Electron.': 'Solid-State Electronics',
  'Solvent Extr. Res. Dev., Jpn.': 'Solvent Extraction Research and Development, Japan',
  'Solvent Extr. Ion Exch.': 'Solvent Extraction and Ion Exchange',
  'S. Afr. J. Chem.': 'South African Journal of Chemistry',
  'Spectrochim. Acta, Part A': 'Spectrochimica Acta, Part A: Molecular and Biomolecular Spectroscopy',
  'Spectrochim. Acta, Part B': 'Spectrochimica Acta, Part B: Atomic Spectroscopy',
  'Spectrosc. Lett.': 'Spectroscopy Letters',
  'Steel Res. Int.': 'Steel Research International',
  'Struct. Chem.': 'Structural Chemistry',
  'Struct. Dyn.': 'Structural Dynamics',
  'Structure (Oxford, U. K.)': 'Structure (Oxford, United Kingdom)',
  'Supercond. Sci. Technol.': 'Superconductor Science and Technology',
  'Superlattices Microstruct.': 'Superlattices and Microstructures',
  'Supramol. Chem.': 'Supramolecular Chemistry',
  'Surf. Eng.': 'Surface Engineering',
  'Surf. Rev. Lett.': 'Surface Review and Letters',
  'Surf. Sci.': 'Surface Science',
  'Surf. Sci. Rep.': 'Surface Science Reports',
  'Surf. Coat. Technol.': 'Surface and Coatings Technology',
  'Surf. Interface Anal.': 'Surface and Interface Analysis',
  'Sustainable Chem. Processes': 'Sustainable Chemical Processes',
  'Synapse (Hoboken, NJ, U. S.)': 'Synapse (Hoboken, NJ, United States)',
  'Synth. React. Inorg., Met.-Org., Nano-Met. Chem.': 'Synthesis and Reactivity in Inorganic, Metal-Organic, and Nano-Metal Chemistry',
  'Synth. Commun.': 'Synthetic Communications',
  'Synth. Met.': 'Synthetic Metals',
  'Tappi J.': 'Tappi Journal',
  'Tech. Phys.': 'Technical Physics',
  'Tech. Phys. Lett.': 'Technical Physics Letters',
  'Tenside, Surfactants, Deterg.': 'Tenside, Surfactants, Detergents',
  'Tetrahedron Lett.': 'Tetrahedron Letters',
  'Text. Res. J.': 'Textile Research Journal',
  'Theor. Chem. Acc.': 'Theoretical Chemistry Accounts',
  'Theor. Found. Chem. Eng.': 'Theoretical Foundations of Chemical Engineering',
  'Theor. Exp. Chem.': 'Theoretical and Experimental Chemistry',
  'Ther. Delivery': 'Therapeutic Delivery',
  'Thermochim. Acta': 'Thermochimica Acta',
  'Thromb. Res.': 'Thrombosis Research',
  'Tonkie Khim. Tekhnol.': 'Tonkie Khimicheskie Tekhnologii',
  'Top. Catal.': 'Topics in Catalysis',
  'Toxicol. Environ. Chem.': 'Toxicological &amp; Environmental Chemistry',
  'Toxicol. Sci.': 'Toxicological Sciences',
  'Toxicol. In Vitro': 'Toxicology In Vitro',
  'Toxicol. Lett.': 'Toxicology Letters',
  'Toxicol. Mech. Methods': 'Toxicology Mechanisms and Methods',
  'Toxicol. Res. (Cambridge, U. K.)': 'Toxicology Research (Cambridge, United Kingdom)',
  'Toxicol. Appl. Pharmacol.': 'Toxicology and Applied Pharmacology',
  'TrAC, Trends Anal. Chem.': 'TrAC, Trends in Analytical Chemistry',
  'Traffic (Oxford, U. K.)': 'Traffic (Oxford, United Kingdom)',
  'Trans. Indian Ceram. Soc.': 'Transactions of the Indian Ceramic Society',
  'Trans. Inst. Met. Finish.': 'Transactions of the Institute of Metal Finishing',
  'Transgenic Res.': 'Transgenic Research',
  'Transition Met. Chem. (Dordrecht, Neth.)': 'Transition Metal Chemistry (Dordrecht, Netherlands)',
  'Trends Biochem. Sci.': 'Trends in Biochemical Sciences',
  'Trends Biotechnol.': 'Trends in Biotechnology',
  'Trends Carbohydr. Res.': 'Trends in Carbohydrate Research',
  'Trends Endocrinol. Metab.': 'Trends in Endocrinology and Metabolism',
  'Trends Environ. Anal. Chem.': 'Trends in Environmental Analytical Chemistry',
  'Trends Food Sci. Technol.': 'Trends in Food Science &amp; Technology',
  'Trends Genet.': 'Trends in Genetics',
  'Trends Immunol.': 'Trends in Immunology',
  'Trends Microbiol.': 'Trends in Microbiology',
  'Trends Neurosci.': 'Trends in Neurosciences',
  'Trends Pharmacol. Sci.': 'Trends in Pharmacological Sciences',
  'Trends Plant Sci.': 'Trends in Plant Science',
  'Tribol. Trans.': 'Tribology Transactions',
  'Turk. J. Chem.': 'Turkish Journal of Chemistry',
  'Ukr. Biochem. J.': 'Ukrainian Biochemical Journal',
  'Ukr. Bioorg. Acta': 'Ukrainica Bioorganica Acta',
  'Ukr. Khim. Zh. (Russ. Ed.)': 'Ukrainskii Khimicheskii Zhurnal (Russian Edition)',
  'Ultrason. Sonochem.': 'Ultrasonics Sonochemistry',
  'Vasc. Pharmacol.': 'Vascular Pharmacology',
  'Vestsi Nats. Akad. Navuk Belarusi, Ser. Khim. Navuk': "Vestsi Natsyyanal'nai Akademii Navuk Belarusi, Seryya Khimichnykh Navuk",
  'Vib. Spectrosc.': 'Vibrational Spectroscopy',
  'Viral Immunol.': 'Viral Immunology',
  'Virus Res.': 'Virus Research',
  'Water Environ. Res.': 'Water Environment Research',
  'Water Qual. Res. J. Can.': 'Water Quality Research Journal of Canada',
  'Water Res.': 'Water Research',
  'Water Resour.': 'Water Resources',
  'Water, Air, Soil Pollut.': 'Water, Air, &amp; Soil Pollution',
  'Wood Sci. Technol.': 'Wood Science and Technology',
  'Wood Fiber Sci.': 'Wood and Fiber Science',
  'World J. Condens. Matter Phys.': 'World Journal of Condensed Matter Physics',
  'X-Ray Spectrom.': 'X-Ray Spectrometry',
  'X-Ray Struct. Anal. Online': 'X-Ray Structure Analysis Online',
  'Z. Anorg. Allg. Chem.': 'Zeitschrift fuer Anorganische und Allgemeine Chemie',
  'Z. Lebensm.-Unters. Forsch. A': 'Zeitschrift fuer Lebensmittel-Untersuchung und -Forschung A: European Food Research and Technology',
  'Z. Naturforsch., A: Phys. Sci.': 'Zeitschrift fuer Naturforschung, A: Physical Sciences',
  'Z. Naturforsch., B: J. Chem. Sci.': 'Zeitschrift fuer Naturforschung, B: A Journal of Chemical Sciences',
  'Z. Naturforsch., C: J. Biosci.': 'Zeitschrift fuer Naturforschung, C: Journal of Biosciences',
  'Z. Phys. Chem. (Muenchen, Ger.)': 'Zeitschrift fuer Physikalische Chemie (Muenchen, Germany)',
  'Zh. Org. Farm. Khim.': 'Zhurnal Organichnoi ta Farmatsevtichnoi Khimii',
}



import subprocess

from .. import utils
from .. import parser
from . common import *


def main(args):
  # check of library path is set
  libraryPath = os.path.expanduser(cfg.get('library_path'))
  if not libraryPath:
    io.err('library_path is not set in config file')
    return

  # set flags that decide what to do
  fulltextSearch = args.find_fulltext
  if fulltextSearch:
    try:
      subprocess.run(['pdf2txt', '-h'], capture_output=True)
    except KeyboardInterrupt:
      raise
    except:
      io.err(f'"pdf2txt -h" failed but is required',
             'for fulltext search, is pdf2txt correctly',
             'installed on your path?')
      return
  search = args.find
  matches = []

  scan = not search and not fulltextSearch
  healthyCount = 0
  foundDuplicates = {}
  foundTrueDuplicates = {}
  unpairedFiles = []
  invalidBibFiles = []
  invalidPdfFiles = []

  # walk through library
  hasWarnedDepth = False
  for root, dirs, files in os.walk(libraryPath, topdown=True):
    # skip annotated folder
    if os.path.relpath(root, libraryPath).startswith('annotated'):
      continue

    # abort tree is too deep
    if root.count(os.sep)-libraryPath.count(os.sep)-2 > cfg.get('max_directory_depth'):
      io.verb(f'skipping subfolders of {root}')
      if not hasWarnedDepth:
        hasWarnedDepth = True
        io.warn(f'reached max_directory_depth='
                f'{cfg.get("max_directory_depth")} when recursing '
                f'library, ignoring deeper levels')
      dirs.clear()

    # skip git directories
    while True:
      i = [d for d in dirs if d.startswith('.git')]
      if not i:
        break
      dirs.remove(i[0])

    # loop through files in dir
    for f in files:
      path = os.path.join(root, f)
      if any([f.lower().endswith('.'+e) for e in cfg.get('bibtex_extensions')]):
        # if bibfile and search is enabled, check of matching
        if search:
          try:
            c = parser.BibFile(path).cites()[0]
          except Exception:
            io.verb(f'skipping invalid library entry {path}')
          else:
            if all([any([s.lower() in v.lower()
                                  for v in c.fields.values()])
                                      for s in search]):
              matches.append(path)

        # if scan is enabled, check if file is successfully parsed
        if scan:
          try:
            c = parser.BibFile(path).cites()[0]
          except Exception:
            invalidBibFiles.append(path)
          else:
            # add to duplicate keys dict
            if c.key not in foundDuplicates:
              foundDuplicates[c.key] = []
            foundDuplicates[c.key].append(path)

            # add to true duplicates dict
            if c.compareAuthorTitle() not in foundTrueDuplicates:
              foundTrueDuplicates[c.compareAuthorTitle()] = []
            foundTrueDuplicates[c.compareAuthorTitle()].append(path)

            if os.path.exists('.'.join(path.split('.')[:-1])+'.pdf'):
              healthyCount += 0.5
            else:
              unpairedFiles.append(path)

      elif f.lower().endswith('.pdf'):
        # if pdf file and fulltext search enabled, check if matching
        if fulltextSearch:
          txtPath = os.path.join(root, '.'+f[:-4]+'.txt')
          # generate txt file if not existing or older than pdf
          if (not os.path.exists(txtPath)
                or os.path.getmtime(txtPath) < os.path.getmtime(path)):
            cmd = ['pdf2txt', '-o', txtPath, path]
            io.verb('running '+' '.join(cmd))
            r = subprocess.run(cmd, capture_output=True)
            if r.returncode:
              io.warn(f'pdf2txt failed with exitcode {r.returncode}:',
                      r.stdout.decode(), r.stderr.decode(), f'skipping library entry {path}')

          if not os.path.isfile(txtPath):
            io.warn(f'pdf2txt did not create text file',
                    f'skipping library entry {path}')
          else:
            with open(txtPath, 'r') as f:
              txt = f.read()
            if all([s.lower() in txt.lower() for s in fulltextSearch]):
              matches.append(path)

        # if scan is enabled, check if file looks valid
        if scan:
          try:
            with open(path, 'rb') as f:
              if not f.read(16).startswith(b'%PDF'):
                raise ValueError('not pdf')
          except KeyboardInterrupt:
            raise
          except:
            invalidPdfFiles.append(path)
          else:
            if any([os.path.exists('.'.join(path.split('.')[:-1])+'.'+e)
                                for e in cfg.get('bibtex_extensions')]):
              healthyCount += 0.5
            else:
              unpairedFiles.append(path)

  if search or fulltextSearch:
    if not matches:
      io.info('no matches')
    for m in sorted(matches):
      bibPath = None
      for ext in cfg.get('bibtex_extensions'):
        _bibPath = utils.replaceSuffix(m, ext)
        if os.path.exists(_bibPath):
          bibPath = _bibPath
          break
      if args.long:
        if bibPath:
          io.raw(parser.BibFile(bibPath).cites()[0].toString())
      elif args.key:
        if bibPath:
          io.info(parser.BibFile(bibPath).cites()[0].key)
      else:
        io.info(m)

  if scan:
    foundDuplicates = {k: v for k, v in foundDuplicates.items()
                                                    if len(v) > 1}
    duplCount = sum([len(v)-1 for v in foundDuplicates.values()])
    foundTrueDuplicates = {k: v for k, v in foundTrueDuplicates.items()
                                                    if len(v) > 1}
    trueDuplCount = sum([len(v)-1 for v in foundTrueDuplicates.values()])
    io.info(f'done scanning library in {libraryPath},',
            f'found {healthyCount} valid looking entries', '-',
            *([f'found {len(unpairedFiles)} unpaired files:']
              +[f'  {f[len(libraryPath)+1:]}' for f in unpairedFiles]+['-']
                        if unpairedFiles else []),

            *([f'found {duplCount} duplicate keys:']
              +['  '+'\n  '.join([_d[len(libraryPath)+1:] for _d in d])
                                      for d in foundDuplicates.values()]+['-']
                        if foundDuplicates else []),

            *([f'found {trueDuplCount} true duplicates (title and authors identical):']
              +['  '+'\n  '.join([_d[len(libraryPath)+1:] for _d in d])
                                      for d in foundTrueDuplicates.values()]+['-']
                        if foundTrueDuplicates else []),

            *([f'found {len(invalidBibFiles)} invalid bib files:']
              +[f'  {b[len(libraryPath)+1:]}' for b in invalidBibFiles]+['-']
                        if invalidBibFiles else []),

            *([f'found {len(invalidPdfFiles)} invalid pdf files:']
              +[f'  {p[len(libraryPath)+1:]}' for p in invalidPdfFiles]
                        if invalidPdfFiles else []))

    # offer removing duplicates
    if (trueDuplCount
        and io.conf(f'remove {trueDuplCount} true duplicates from library?',
                    default=False)):
      for paths in foundTrueDuplicates.values():
        for path in sorted(paths)[:-1]:
          dir, base = os.path.split(path)
          base = '.'.join(base.split('.')[:-1])
          for p in os.listdir(dir):
            if p.startswith(base):
              rem = os.path.join(dir, p)
              io.info(f'removing {rem}')
              os.remove(rem)
